-module(core_parse).
-define(THIS_MODULE, core_parse).
-export([parse/1, parse_and_scan/1, format_error/1]).

-export([abstract/1,abstract/2,normalise/1]).

-include("core_parse.hrl").

tok_val(T) -> element(3, T).
tok_line(T) -> element(2, T).

abstract(T, _N) -> abstract(T).

abstract(Term) -> core_lib:make_literal(Term).

normalise(Core) -> core_lib:literal_value(Core).

%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The parser generator will insert appropriate declarations before this line.%

parse(Tokens) ->
    case catch yeccpars1(Tokens, false, 0, [], []) of
	error ->
	    Errorline =
		if Tokens == [] -> 0; true -> element(2, hd(Tokens)) end,
	    {error,
	     {Errorline, ?THIS_MODULE, "syntax error at or after this line."}};
	Other ->
	    Other
    end.

parse_and_scan({Mod, Fun, Args}) ->
    case apply(Mod, Fun, Args) of
	{eof, _} ->
	    {ok, eof};
	{error, Descriptor, _} ->
	    {error, Descriptor};
	{ok, Tokens, _} ->
	    yeccpars1(Tokens, {Mod, Fun, Args}, 0, [], [])
    end.

format_error(Message) ->
    case io_lib:deep_char_list(Message) of
	true ->
	    Message;
	_ ->
	    io_lib:write(Message)
    end.

% To be used in grammar files to throw an error message to the parser toplevel.
% Doesn't have to be exported!
return_error(Line, Message) ->
    throw({error, {Line, ?THIS_MODULE, Message}}).


% Don't change yeccpars1/6 too much, it is called recursively by yeccpars2/8!
yeccpars1([Token | Tokens], Tokenizer, State, States, Vstack) ->
    yeccpars2(State, element(1, Token), States, Vstack, Token, Tokens,
	      Tokenizer);
yeccpars1([], {M, F, A}, State, States, Vstack) ->
    case catch apply(M, F, A) of
        {eof, Endline} ->
            {error, {Endline, ?THIS_MODULE, "end_of_file"}};
        {error, Descriptor, _Endline} ->
            {error, Descriptor};
        {'EXIT', Reason} ->
            {error, {0, ?THIS_MODULE, Reason}};
        {ok, Tokens, _Endline} ->
	    case catch yeccpars1(Tokens, {M, F, A}, State, States, Vstack) of
		error ->
		    Errorline = element(2, hd(Tokens)),
		    {error, {Errorline, ?THIS_MODULE,
			     "syntax error at or after this line."}};
		Other ->
		    Other
	    end
    end;
yeccpars1([], false, State, States, Vstack) ->
    yeccpars2(State, '$end', States, Vstack, {'$end', 999999}, [], false).

% For internal use only.
yeccerror(Token) ->
    {error,
     {element(2, Token), ?THIS_MODULE,
      ["syntax error before: ", yecctoken2string(Token)]}}.

yecctoken2string({atom, _, A}) -> io_lib:write(A);
yecctoken2string({integer,_,N}) -> io_lib:write(N);
yecctoken2string({float,_,F}) -> io_lib:write(F);
yecctoken2string({char,_,C}) -> io_lib:write_char(C);
yecctoken2string({var,_,V}) -> io_lib:format('~s', [V]);
yecctoken2string({string,_,S}) -> io_lib:write_string(S);
yecctoken2string({reserved_symbol, _, A}) -> io_lib:format('~w', [A]);
yecctoken2string({_Cat, _, Val}) -> io_lib:format('~w', [Val]);

yecctoken2string({'dot', _}) -> io_lib:format('~w', ['.']);
yecctoken2string({'$end', _}) ->
    [];
yecctoken2string({Other, _}) when atom(Other) ->
    io_lib:format('~w', [Other]);
yecctoken2string(Other) ->
    io_lib:write(Other).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


yeccpars2(0, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, 'module', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 2, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(1, 'module', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 319, [1 | __Ss], [__T | __Stack]);
yeccpars2(1, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(2, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 4, [2 | __Ss], [__T | __Stack]);
yeccpars2(2, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(3, '$end', _, __Stack, _, _, _) ->
 {ok, hd(__Stack)};
yeccpars2(3, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(4, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 5, [4 | __Ss], [__T | __Stack]);
yeccpars2(4, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(5, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 312, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(6, 'attributes', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [6 | __Ss], [__T | __Stack]);
yeccpars2(6, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(7, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 282, [7 | __Ss], [__T | __Stack]);
yeccpars2(7, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(8, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [8 | __Ss], [__T | __Stack]);
yeccpars2(8, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [8 | __Ss], [__T | __Stack]);
yeccpars2(8, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(13, __Cat, [8 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(9, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [9 | __Ss], [__T | __Stack]);
yeccpars2(9, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(10, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [10 | __Ss], [__T | __Stack]);
yeccpars2(10, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(11, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [11 | __Ss], [__T | __Stack]);
yeccpars2(11, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(12, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [12 | __Ss], [__T | __Stack]);
yeccpars2(12, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [12 | __Ss], [__T | __Stack]);
yeccpars2(12, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(17, __Cat, [12 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(13, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(module_defs, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(14, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_function_name, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(15, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [15 | __Ss], [__T | __Stack]);
yeccpars2(15, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(16, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_module{name = #c_atom{val = tok_val(__2)}, exports = __3, attrs = __4, defs = __5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(module_definition, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(17, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(function_definitions, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(18, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [18 | __Ss], [__T | __Stack]);
yeccpars2(18, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(19, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_fname{id = tok_val(__1), arity = tok_val(__3)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(function_name, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(20, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [20 | __Ss], [__T | __Stack]);
yeccpars2(20, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(21, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [21 | __Ss], [__T | __Stack]);
yeccpars2(21, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(22, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_def{name = __1, val = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(function_definition, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(23, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [23 | __Ss], [__T | __Stack]);
yeccpars2(23, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(24, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_fun, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(25, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [25 | __Ss], [__T | __Stack]);
yeccpars2(25, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(26, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [26 | __Ss], [__T | __Stack]);
yeccpars2(26, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(27, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 271, [27 | __Ss], [__T | __Stack]);
yeccpars2(27, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(28, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 269, [28 | __Ss], [__T | __Stack]);
yeccpars2(28, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(anno_variables, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(29, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [29 | __Ss], [__T | __Stack]);
yeccpars2(29, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(30, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_var{name = tok_val(__1)},
 yeccpars2(yeccgoto(variable, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(31, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_variable, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(32, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [32 | __Ss], [__T | __Stack]);
yeccpars2(32, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(33, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [33 | __Ss], [__T | __Stack]);
yeccpars2(33, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(34, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, '>#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 254, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [34 | __Ss], [__T | __Stack]);
yeccpars2(34, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(35, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [35 | __Ss], [__T | __Stack]);
yeccpars2(35, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(36, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 247, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [36 | __Ss], [__T | __Stack]);
yeccpars2(36, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(37, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 149, [37 | __Ss], [__T | __Stack]);
yeccpars2(37, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(38, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_fun{vars = __3, body = __6},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(fun_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(39, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(40, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [40 | __Ss], [__T | __Stack]);
yeccpars2(40, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(41, '/', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [41 | __Ss], [__T | __Stack]);
yeccpars2(41, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_atom{val = tok_val(__1)},
 yeccpars2(yeccgoto(atomic_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(42, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(43, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(44, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [44 | __Ss], [__T | __Stack]);
yeccpars2(44, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(45, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(46, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [46 | __Ss], [__T | __Stack]);
yeccpars2(46, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(47, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(48, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [48 | __Ss], [__T | __Stack]);
yeccpars2(48, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(49, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(50, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_char{val = tok_val(__1)},
 yeccpars2(yeccgoto(atomic_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(51, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(52, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [52 | __Ss], [__T | __Stack]);
yeccpars2(52, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(53, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(54, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_float{val = tok_val(__1)},
 yeccpars2(yeccgoto(atomic_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(55, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(56, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(57, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_int{val = tok_val(__1)},
 yeccpars2(yeccgoto(atomic_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(58, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [58 | __Ss], [__T | __Stack]);
yeccpars2(58, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(59, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(60, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [60 | __Ss], [__T | __Stack]);
yeccpars2(60, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(217, __Cat, [60 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(61, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(62, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_nil{},
 yeccpars2(yeccgoto(atomic_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(63, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 211, [63 | __Ss], [__T | __Stack]);
yeccpars2(63, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(64, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(65, 'after', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 99, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [65 | __Ss], [__T | __Stack]);
yeccpars2(65, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(66, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(67, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(68, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(69, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_string{val = tok_val(__1)},
 yeccpars2(yeccgoto(atomic_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(70, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [70 | __Ss], [__T | __Stack]);
yeccpars2(70, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(71, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(72, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(73, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(single_expression, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(74, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [74 | __Ss], [__T | __Stack]);
yeccpars2(74, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(75, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [75 | __Ss], [__T | __Stack]);
yeccpars2(75, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(anno_expressions, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(76, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [76 | __Ss], [__T | __Stack]);
yeccpars2(76, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(77, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_tuple{es = []},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(78, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_tuple{es = __2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tuple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(79, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [79 | __Ss], [__T | __Stack]);
yeccpars2(79, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(80, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(anno_expressions, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(81, 'of', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [81 | __Ss], [__T | __Stack]);
yeccpars2(81, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(82, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [82 | __Ss], [__T | __Stack]);
yeccpars2(82, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(83, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 92, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [83 | __Ss], [__T | __Stack]);
yeccpars2(83, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(84, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(let_vars, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(85, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [85 | __Ss], [__T | __Stack]);
yeccpars2(85, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(86, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [86 | __Ss], [__T | __Stack]);
yeccpars2(86, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(87, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [87 | __Ss], [__T | __Stack]);
yeccpars2(87, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(88, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [88 | __Ss], [__T | __Stack]);
yeccpars2(88, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(89, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [89 | __Ss], [__T | __Stack]);
yeccpars2(89, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(90, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [90 | __Ss], [__T | __Stack]);
yeccpars2(90, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(91, __Cat, __Ss,  [__10,__9,__8,__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  if length(__8) == 2 -> #c_try{arg = __2, vars = __4, body = __6, evars = __8, handler = __10}; true -> return_error(tok_line(__7),"syntax error after catch") end,
 __Nss = lists:nthtail(9, __Ss),
 yeccpars2(yeccgoto(try_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(92, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(let_vars, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(93, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(94, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(let_vars, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(95, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '>#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 190, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [95 | __Ss], [__T | __Stack]);
yeccpars2(95, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(96, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [96 | __Ss], [__T | __Stack]);
yeccpars2(96, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(97, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 182, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(98, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 149, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(99, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [99 | __Ss], [__T | __Stack]);
yeccpars2(99, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(100, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(anno_clauses, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(101, 'after', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 99, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(102, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(clause_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(103, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 162, [103 | __Ss], [__T | __Stack]);
yeccpars2(103, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(104, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_atom{val = tok_val(__1)},
 yeccpars2(yeccgoto(atomic_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(105, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(atomic_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(106, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(other_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(107, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(other_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(108, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_clause, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(109, 'when', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(110, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(other_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(111, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(112, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {T,A} = __2, #c_receive{clauses = [], timeout = T, action = A}
  end,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(113, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(other_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(114, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 118, [114 | __Ss], [__T | __Stack]);
yeccpars2(114, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(115, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [115 | __Ss], [__T | __Stack]);
yeccpars2(115, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(116, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 120, [116 | __Ss], [__T | __Stack]);
yeccpars2(116, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(anno_patterns, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(117, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 119, [117 | __Ss], [__T | __Stack]);
yeccpars2(117, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(118, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_tuple{es = []},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(tuple_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(119, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_tuple{es = __2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tuple_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(120, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(121, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(anno_patterns, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(122, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 162, [122 | __Ss], [__T | __Stack]);
yeccpars2(122, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(123, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 159, [123 | __Ss], [__T | __Stack]);
yeccpars2(123, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(124, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_variable, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(125, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(126, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 129, [126 | __Ss], [__T | __Stack]);
yeccpars2(126, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [126 | __Ss], [__T | __Stack]);
yeccpars2(126, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [126 | __Ss], [__T | __Stack]);
yeccpars2(126, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [126 | __Ss], [__T | __Stack]);
yeccpars2(126, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [126 | __Ss], [__T | __Stack]);
yeccpars2(126, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [126 | __Ss], [__T | __Stack]);
yeccpars2(126, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [126 | __Ss], [__T | __Stack]);
yeccpars2(126, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 130, [126 | __Ss], [__T | __Stack]);
yeccpars2(126, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(127, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 128, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(128, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  core_lib:set_anno(__2,__4),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(anno_variable, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(129, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 129, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 149, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(130, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(annotation, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(131, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  tok_val(__1),
 yeccpars2(yeccgoto(atomic_constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(132, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(133, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  tok_val(__1),
 yeccpars2(yeccgoto(atomic_constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(134, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(135, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 147, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(constants, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(136, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 146, [136 | __Ss], [__T | __Stack]);
yeccpars2(136, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(137, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  tok_val(__1),
 yeccpars2(yeccgoto(atomic_constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(138, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  tok_val(__1),
 yeccpars2(yeccgoto(atomic_constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(139, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(yeccgoto(atomic_constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(140, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  tok_val(__1),
 yeccpars2(yeccgoto(atomic_constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(141, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(142, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 129, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(143, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 145, [143 | __Ss], [__T | __Stack]);
yeccpars2(143, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(144, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(tuple_constant, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(145, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  list_to_tuple(__2),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tuple_constant, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(146, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(annotation, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(147, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 129, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(148, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(constants, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(149, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {nil,tok_line(__1)},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(nil, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(150, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 151, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 154, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 152, [150 | __Ss], [__T | __Stack]);
yeccpars2(150, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(151, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 129, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [151 | __Ss], [__T | __Stack]);
yeccpars2(151, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(152, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(yeccgoto(tail_constant, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(153, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(cons_constant, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(154, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 129, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 131, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [154 | __Ss], [__T | __Stack]);
yeccpars2(154, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(155, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 156, [155 | __Ss], [__T | __Stack]);
yeccpars2(155, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(156, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail_constant, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(157, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 151, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 154, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 152, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(158, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail_constant, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(159, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(160, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 161, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(161, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  core_lib:set_anno(__2,__4),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(anno_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(162, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(163, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_alias{var = __1, pat = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(other_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(164, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(165, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(166, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(167, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_clause{pats = __1, guard = __3, body = __5},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(168, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {T,A} = __3, #c_receive{clauses = __2, timeout = T, action = A}
  end,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(receive_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(169, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(anno_clauses, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(170, '->', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 171, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(171, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [171 | __Ss], [__T | __Stack]);
yeccpars2(171, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(172, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__2,__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(timeout, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(173, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 174, [173 | __Ss], [__T | __Stack]);
yeccpars2(173, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 177, [173 | __Ss], [__T | __Stack]);
yeccpars2(173, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 175, [173 | __Ss], [__T | __Stack]);
yeccpars2(173, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(174, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [174 | __Ss], [__T | __Stack]);
yeccpars2(174, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [174 | __Ss], [__T | __Stack]);
yeccpars2(174, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [174 | __Ss], [__T | __Stack]);
yeccpars2(174, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [174 | __Ss], [__T | __Stack]);
yeccpars2(174, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [174 | __Ss], [__T | __Stack]);
yeccpars2(174, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [174 | __Ss], [__T | __Stack]);
yeccpars2(174, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [174 | __Ss], [__T | __Stack]);
yeccpars2(174, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [174 | __Ss], [__T | __Stack]);
yeccpars2(174, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [174 | __Ss], [__T | __Stack]);
yeccpars2(174, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [174 | __Ss], [__T | __Stack]);
yeccpars2(174, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(175, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_nil{},
 yeccpars2(yeccgoto(tail_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(176, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_cons{hd = __2, tl = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(cons_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(177, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [177 | __Ss], [__T | __Stack]);
yeccpars2(177, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [177 | __Ss], [__T | __Stack]);
yeccpars2(177, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [177 | __Ss], [__T | __Stack]);
yeccpars2(177, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [177 | __Ss], [__T | __Stack]);
yeccpars2(177, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [177 | __Ss], [__T | __Stack]);
yeccpars2(177, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 115, [177 | __Ss], [__T | __Stack]);
yeccpars2(177, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [177 | __Ss], [__T | __Stack]);
yeccpars2(177, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [177 | __Ss], [__T | __Stack]);
yeccpars2(177, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [177 | __Ss], [__T | __Stack]);
yeccpars2(177, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [177 | __Ss], [__T | __Stack]);
yeccpars2(177, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(178, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 179, [178 | __Ss], [__T | __Stack]);
yeccpars2(178, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(179, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(180, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 174, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 177, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 175, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(181, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_cons{hd = __2, tl = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(182, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(clause_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(183, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 184, [183 | __Ss], [__T | __Stack]);
yeccpars2(183, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(184, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(clause_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(185, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 187, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(186, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 159, [186 | __Ss], [__T | __Stack]);
yeccpars2(186, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(anno_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(187, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [187 | __Ss], [__T | __Stack]);
yeccpars2(187, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(188, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 189, [188 | __Ss], [__T | __Stack]);
yeccpars2(188, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(189, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  core_lib:set_anno(__2,__4),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(anno_clause, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(190, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_binary{segs = []},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(binary_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(191, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(192, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 199, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(193, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 197, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(segment_patterns, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(194, '>#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 196, [194 | __Ss], [__T | __Stack]);
yeccpars2(194, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(195, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars2(yeccgoto(anno_variable, hd(__Ss)), '=', __Ss, __Stack, __T, __Ts, __Tzr);
yeccpars2(195, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(196, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_binary{segs = __2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(binary_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(197, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(198, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(segment_patterns, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(199, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [199 | __Ss], [__T | __Stack]);
yeccpars2(199, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(200, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 209, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,1},
 yeccpars2(yeccgoto(seg_size_pattern, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(201, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 202, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(202, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 203, [202 | __Ss], [__T | __Stack]);
yeccpars2(202, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(203, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  tok_val(__1),
 yeccpars2(yeccgoto(seg_type, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(204, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 205, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(206, __Cat, [204 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(205, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 207, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(206, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {Size,Unit} = __3, #c_bin_seg{val = __1, size = Size, unit = Unit, type = __5, flags = __6}
  end,
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(segment_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(207, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 205, [207 | __Ss], [__T | __Stack]);
yeccpars2(207, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(208, __Cat, [207 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(208, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [tok_val(__2)|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(seg_flags, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(209, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 210, [209 | __Ss], [__T | __Stack]);
yeccpars2(209, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(210, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,tok_val(__3)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(seg_size_pattern, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(211, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 212, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(212, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 214, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(213, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   Name = #c_atom{val = tok_val(__2)}, #c_primop{name = Name, args = __3}
  end,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(primop_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(214, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(arg_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(215, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 216, [215 | __Ss], [__T | __Stack]);
yeccpars2(215, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(216, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(arg_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(217, 'in', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 218, [217 | __Ss], [__T | __Stack]);
yeccpars2(217, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(218, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [218 | __Ss], [__T | __Stack]);
yeccpars2(218, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(219, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_letrec{defs = __2, body = __4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(letrec_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(220, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 221, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(221, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(222, 'in', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 223, [222 | __Ss], [__T | __Stack]);
yeccpars2(222, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(223, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [223 | __Ss], [__T | __Stack]);
yeccpars2(223, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(224, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_let{vars = __2, arg = __4, body = __6},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(let_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(225, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(226, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_seq{arg = __2, body = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(sequence, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(227, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_catch{body = __2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(catch_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(228, 'of', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 229, [228 | __Ss], [__T | __Stack]);
yeccpars2(228, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(229, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 97, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 96, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 95, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 98, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 114, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(230, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 231, [230 | __Ss], [__T | __Stack]);
yeccpars2(230, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(231, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_case{arg = __2, clauses = __4},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(case_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(232, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 233, [232 | __Ss], [__T | __Stack]);
yeccpars2(232, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(233, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(234, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 212, [234 | __Ss], [__T | __Stack]);
yeccpars2(234, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(235, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_call{module = __2, name = __4, args = __5},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(call_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(236, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 212, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(237, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_apply{op = __2, args = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(application_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(238, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 239, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 242, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 240, [238 | __Ss], [__T | __Stack]);
yeccpars2(238, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(239, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(240, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_nil{},
 yeccpars2(yeccgoto(tail, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(241, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_cons{hd = __2, tl = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(cons, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(242, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [242 | __Ss], [__T | __Stack]);
yeccpars2(242, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(243, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 244, [243 | __Ss], [__T | __Stack]);
yeccpars2(243, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(244, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(245, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 239, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 242, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 240, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(246, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_cons{hd = __2, tl = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(247, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_values{es = []},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(expression, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(248, '>', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 249, [248 | __Ss], [__T | __Stack]);
yeccpars2(248, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(249, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_values{es = __2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(expression, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(250, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 251, [250 | __Ss], [__T | __Stack]);
yeccpars2(250, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(251, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(252, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 253, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(253, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  core_lib:set_anno(__2,__4),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(anno_expression, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(254, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_binary{segs = []},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(binary, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(255, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 267, [255 | __Ss], [__T | __Stack]);
yeccpars2(255, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(segments, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(256, '>#', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 266, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(257, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 258, [257 | __Ss], [__T | __Stack]);
yeccpars2(257, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(258, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(259, ':', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 263, [259 | __Ss], [__T | __Stack]);
yeccpars2(259, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(260, '*', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 261, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,1},
 yeccpars2(yeccgoto(seg_size_unit, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(261, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 262, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(262, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,tok_val(__3)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(seg_size_unit, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(263, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 203, [263 | __Ss], [__T | __Stack]);
yeccpars2(263, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(264, '-', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 205, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(265, __Cat, [264 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(265, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val = begin
   {Size,Unit} = __3, #c_bin_seg{val = __1, size = Size, unit = Unit, type = __5, flags = __6}
  end,
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(segment, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(266, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_binary{segs = __2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(binary, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(267, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(268, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(segments, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(269, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(270, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(anno_variables, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(271, 'receive', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'catch', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'try', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'primop', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'call', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'apply', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'case', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'letrec', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'let', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'fun', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'do', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, '#<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, '<', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, 'var', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [271 | __Ss], [__T | __Stack]);
yeccpars2(271, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(272, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_fun{vars = [], body = __5},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(fun_expr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(273, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 125, [273 | __Ss], [__T | __Stack]);
yeccpars2(273, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(274, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 275, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(275, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(276, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 277, [276 | __Ss], [__T | __Stack]);
yeccpars2(276, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(277, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  core_lib:set_anno(__2,__4),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(anno_fun, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(278, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 279, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(279, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [279 | __Ss], [__T | __Stack]);
yeccpars2(279, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(280, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 281, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(281, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  core_lib:set_anno(__2,__4),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(anno_function_name, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(282, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 284, [282 | __Ss], [__T | __Stack]);
yeccpars2(282, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 283, [282 | __Ss], [__T | __Stack]);
yeccpars2(282, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(283, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(module_attribute, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(284, '=', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 290, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(285, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 288, [285 | __Ss], [__T | __Stack]);
yeccpars2(285, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(attribute_list, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(286, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 287, [286 | __Ss], [__T | __Stack]);
yeccpars2(286, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(287, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(module_attribute, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(288, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 284, [288 | __Ss], [__T | __Stack]);
yeccpars2(288, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(289, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(attribute_list, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(290, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 291, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 296, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(291, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 291, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 296, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 149, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(292, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(293, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(294, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_def{name = #c_atom{val = tok_val(__1)}, val = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(attribute, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(295, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(296, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 291, [296 | __Ss], [__T | __Stack]);
yeccpars2(296, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 296, [296 | __Ss], [__T | __Stack]);
yeccpars2(296, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 299, [296 | __Ss], [__T | __Stack]);
yeccpars2(296, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [296 | __Ss], [__T | __Stack]);
yeccpars2(296, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [296 | __Ss], [__T | __Stack]);
yeccpars2(296, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [296 | __Ss], [__T | __Stack]);
yeccpars2(296, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [296 | __Ss], [__T | __Stack]);
yeccpars2(296, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [296 | __Ss], [__T | __Stack]);
yeccpars2(296, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(297, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 301, [297 | __Ss], [__T | __Stack]);
yeccpars2(297, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(literals, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(298, '}', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 300, [298 | __Ss], [__T | __Stack]);
yeccpars2(298, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(299, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_tuple{es = []},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(tuple_literal, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(300, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_tuple{es = __2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tuple_literal, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(301, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 291, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 296, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(302, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(literals, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(303, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 304, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 307, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 305, [303 | __Ss], [__T | __Stack]);
yeccpars2(303, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(304, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 291, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 296, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [304 | __Ss], [__T | __Stack]);
yeccpars2(304, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(305, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_nil{},
 yeccpars2(yeccgoto(tail_literal, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(306, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_cons{hd = __2, tl = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(cons_literal, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(307, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 291, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, '{', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 296, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'string', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 104, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'float', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'integer', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, 'char', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [307 | __Ss], [__T | __Stack]);
yeccpars2(307, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(308, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 309, [308 | __Ss], [__T | __Stack]);
yeccpars2(308, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(309, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail_literal, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(310, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 304, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, '|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 307, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 305, [310 | __Ss], [__T | __Stack]);
yeccpars2(310, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(311, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_cons{hd = __2, tl = __3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(tail_literal, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(312, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(module_export, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(313, ',', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 317, [313 | __Ss], [__T | __Stack]);
yeccpars2(313, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(exported_names, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(314, ']', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 316, [314 | __Ss], [__T | __Stack]);
yeccpars2(314, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(315, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(exported_name, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(316, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(module_export, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(317, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(318, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(exported_names, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(319, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 320, [319 | __Ss], [__T | __Stack]);
yeccpars2(319, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(320, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 5, [320 | __Ss], [__T | __Stack]);
yeccpars2(320, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(321, 'attributes', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [321 | __Ss], [__T | __Stack]);
yeccpars2(321, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(322, '(', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [322 | __Ss], [__T | __Stack]);
yeccpars2(322, 'atom', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [322 | __Ss], [__T | __Stack]);
yeccpars2(322, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(13, __Cat, [322 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(323, 'end', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 324, [323 | __Ss], [__T | __Stack]);
yeccpars2(323, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(324, '-|', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 325, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(325, '[', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [325 | __Ss], [__T | __Stack]);
yeccpars2(325, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(326, ')', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 327, [326 | __Ss], [__T | __Stack]);
yeccpars2(326, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(327, __Cat, __Ss,  [__10,__9,__8,__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #c_module{anno = __9, name = tok_val(__3), exports = __4, attrs = __5, defs = __6},
 __Nss = lists:nthtail(9, __Ss),
 yeccpars2(yeccgoto(module_definition, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(__Other, _, _, _, _, _, _) ->
 exit({parser, __Other, missing_state_in_action_table}).

yeccgoto(anno_clause, 65) ->
 100;
yeccgoto(anno_clause, 100) ->
 100;
yeccgoto(anno_clause, 229) ->
 100;
yeccgoto(anno_clauses, 65) ->
 101;
yeccgoto(anno_clauses, 100) ->
 169;
yeccgoto(anno_clauses, 229) ->
 230;
yeccgoto(anno_expression, 33) ->
 38;
yeccgoto(anno_expression, 36) ->
 75;
yeccgoto(anno_expression, 37) ->
 238;
yeccgoto(anno_expression, 40) ->
 236;
yeccgoto(anno_expression, 44) ->
 232;
yeccgoto(anno_expression, 46) ->
 228;
yeccgoto(anno_expression, 48) ->
 227;
yeccgoto(anno_expression, 52) ->
 225;
yeccgoto(anno_expression, 70) ->
 81;
yeccgoto(anno_expression, 74) ->
 75;
yeccgoto(anno_expression, 79) ->
 75;
yeccgoto(anno_expression, 86) ->
 87;
yeccgoto(anno_expression, 90) ->
 91;
yeccgoto(anno_expression, 99) ->
 170;
yeccgoto(anno_expression, 164) ->
 165;
yeccgoto(anno_expression, 166) ->
 167;
yeccgoto(anno_expression, 171) ->
 172;
yeccgoto(anno_expression, 212) ->
 75;
yeccgoto(anno_expression, 218) ->
 219;
yeccgoto(anno_expression, 221) ->
 222;
yeccgoto(anno_expression, 223) ->
 224;
yeccgoto(anno_expression, 225) ->
 226;
yeccgoto(anno_expression, 233) ->
 234;
yeccgoto(anno_expression, 239) ->
 245;
yeccgoto(anno_expression, 242) ->
 243;
yeccgoto(anno_expression, 271) ->
 272;
yeccgoto(anno_expressions, 36) ->
 248;
yeccgoto(anno_expressions, 74) ->
 76;
yeccgoto(anno_expressions, 79) ->
 80;
yeccgoto(anno_expressions, 212) ->
 215;
yeccgoto(anno_fun, 20) ->
 22;
yeccgoto(anno_function_name, 8) ->
 10;
yeccgoto(anno_function_name, 12) ->
 10;
yeccgoto(anno_function_name, 60) ->
 10;
yeccgoto(anno_function_name, 322) ->
 10;
yeccgoto(anno_pattern, 65) ->
 102;
yeccgoto(anno_pattern, 96) ->
 102;
yeccgoto(anno_pattern, 97) ->
 116;
yeccgoto(anno_pattern, 98) ->
 173;
yeccgoto(anno_pattern, 100) ->
 102;
yeccgoto(anno_pattern, 114) ->
 116;
yeccgoto(anno_pattern, 120) ->
 116;
yeccgoto(anno_pattern, 162) ->
 163;
yeccgoto(anno_pattern, 174) ->
 180;
yeccgoto(anno_pattern, 177) ->
 178;
yeccgoto(anno_pattern, 229) ->
 102;
yeccgoto(anno_patterns, 97) ->
 183;
yeccgoto(anno_patterns, 114) ->
 117;
yeccgoto(anno_patterns, 120) ->
 121;
yeccgoto(anno_variable, 25) ->
 28;
yeccgoto(anno_variable, 58) ->
 84;
yeccgoto(anno_variable, 65) ->
 103;
yeccgoto(anno_variable, 82) ->
 84;
yeccgoto(anno_variable, 83) ->
 28;
yeccgoto(anno_variable, 88) ->
 84;
yeccgoto(anno_variable, 95) ->
 122;
yeccgoto(anno_variable, 96) ->
 103;
yeccgoto(anno_variable, 97) ->
 103;
yeccgoto(anno_variable, 98) ->
 103;
yeccgoto(anno_variable, 100) ->
 103;
yeccgoto(anno_variable, 114) ->
 103;
yeccgoto(anno_variable, 115) ->
 122;
yeccgoto(anno_variable, 120) ->
 103;
yeccgoto(anno_variable, 162) ->
 103;
yeccgoto(anno_variable, 174) ->
 103;
yeccgoto(anno_variable, 177) ->
 103;
yeccgoto(anno_variable, 197) ->
 122;
yeccgoto(anno_variable, 199) ->
 122;
yeccgoto(anno_variable, 229) ->
 103;
yeccgoto(anno_variable, 269) ->
 28;
yeccgoto(anno_variables, 25) ->
 29;
yeccgoto(anno_variables, 83) ->
 93;
yeccgoto(anno_variables, 269) ->
 270;
yeccgoto(annotation, 125) ->
 127;
yeccgoto(annotation, 159) ->
 160;
yeccgoto(annotation, 187) ->
 188;
yeccgoto(annotation, 251) ->
 252;
yeccgoto(annotation, 275) ->
 276;
yeccgoto(annotation, 279) ->
 280;
yeccgoto(annotation, 325) ->
 326;
yeccgoto(application_expr, 33) ->
 39;
yeccgoto(application_expr, 34) ->
 39;
yeccgoto(application_expr, 35) ->
 39;
yeccgoto(application_expr, 36) ->
 39;
yeccgoto(application_expr, 37) ->
 39;
yeccgoto(application_expr, 40) ->
 39;
yeccgoto(application_expr, 44) ->
 39;
yeccgoto(application_expr, 46) ->
 39;
yeccgoto(application_expr, 48) ->
 39;
yeccgoto(application_expr, 52) ->
 39;
yeccgoto(application_expr, 70) ->
 39;
yeccgoto(application_expr, 74) ->
 39;
yeccgoto(application_expr, 79) ->
 39;
yeccgoto(application_expr, 86) ->
 39;
yeccgoto(application_expr, 90) ->
 39;
yeccgoto(application_expr, 99) ->
 39;
yeccgoto(application_expr, 164) ->
 39;
yeccgoto(application_expr, 166) ->
 39;
yeccgoto(application_expr, 171) ->
 39;
yeccgoto(application_expr, 212) ->
 39;
yeccgoto(application_expr, 218) ->
 39;
yeccgoto(application_expr, 221) ->
 39;
yeccgoto(application_expr, 223) ->
 39;
yeccgoto(application_expr, 225) ->
 39;
yeccgoto(application_expr, 233) ->
 39;
yeccgoto(application_expr, 239) ->
 39;
yeccgoto(application_expr, 242) ->
 39;
yeccgoto(application_expr, 258) ->
 39;
yeccgoto(application_expr, 267) ->
 39;
yeccgoto(application_expr, 271) ->
 39;
yeccgoto(arg_list, 211) ->
 213;
yeccgoto(arg_list, 234) ->
 235;
yeccgoto(arg_list, 236) ->
 237;
yeccgoto(atomic_constant, 126) ->
 132;
yeccgoto(atomic_constant, 129) ->
 132;
yeccgoto(atomic_constant, 142) ->
 132;
yeccgoto(atomic_constant, 147) ->
 132;
yeccgoto(atomic_constant, 151) ->
 132;
yeccgoto(atomic_constant, 154) ->
 132;
yeccgoto(atomic_literal, 33) ->
 42;
yeccgoto(atomic_literal, 34) ->
 42;
yeccgoto(atomic_literal, 35) ->
 42;
yeccgoto(atomic_literal, 36) ->
 42;
yeccgoto(atomic_literal, 37) ->
 42;
yeccgoto(atomic_literal, 40) ->
 42;
yeccgoto(atomic_literal, 44) ->
 42;
yeccgoto(atomic_literal, 46) ->
 42;
yeccgoto(atomic_literal, 48) ->
 42;
yeccgoto(atomic_literal, 52) ->
 42;
yeccgoto(atomic_literal, 65) ->
 105;
yeccgoto(atomic_literal, 70) ->
 42;
yeccgoto(atomic_literal, 74) ->
 42;
yeccgoto(atomic_literal, 79) ->
 42;
yeccgoto(atomic_literal, 86) ->
 42;
yeccgoto(atomic_literal, 90) ->
 42;
yeccgoto(atomic_literal, 95) ->
 105;
yeccgoto(atomic_literal, 96) ->
 105;
yeccgoto(atomic_literal, 97) ->
 105;
yeccgoto(atomic_literal, 98) ->
 105;
yeccgoto(atomic_literal, 99) ->
 42;
yeccgoto(atomic_literal, 100) ->
 105;
yeccgoto(atomic_literal, 114) ->
 105;
yeccgoto(atomic_literal, 115) ->
 105;
yeccgoto(atomic_literal, 120) ->
 105;
yeccgoto(atomic_literal, 162) ->
 105;
yeccgoto(atomic_literal, 164) ->
 42;
yeccgoto(atomic_literal, 166) ->
 42;
yeccgoto(atomic_literal, 171) ->
 42;
yeccgoto(atomic_literal, 174) ->
 105;
yeccgoto(atomic_literal, 177) ->
 105;
yeccgoto(atomic_literal, 197) ->
 105;
yeccgoto(atomic_literal, 199) ->
 105;
yeccgoto(atomic_literal, 212) ->
 42;
yeccgoto(atomic_literal, 218) ->
 42;
yeccgoto(atomic_literal, 221) ->
 42;
yeccgoto(atomic_literal, 223) ->
 42;
yeccgoto(atomic_literal, 225) ->
 42;
yeccgoto(atomic_literal, 229) ->
 105;
yeccgoto(atomic_literal, 233) ->
 42;
yeccgoto(atomic_literal, 239) ->
 42;
yeccgoto(atomic_literal, 242) ->
 42;
yeccgoto(atomic_literal, 258) ->
 42;
yeccgoto(atomic_literal, 267) ->
 42;
yeccgoto(atomic_literal, 271) ->
 42;
yeccgoto(atomic_literal, 290) ->
 292;
yeccgoto(atomic_literal, 291) ->
 292;
yeccgoto(atomic_literal, 296) ->
 292;
yeccgoto(atomic_literal, 301) ->
 292;
yeccgoto(atomic_literal, 304) ->
 292;
yeccgoto(atomic_literal, 307) ->
 292;
yeccgoto(atomic_pattern, 65) ->
 106;
yeccgoto(atomic_pattern, 95) ->
 106;
yeccgoto(atomic_pattern, 96) ->
 106;
yeccgoto(atomic_pattern, 97) ->
 106;
yeccgoto(atomic_pattern, 98) ->
 106;
yeccgoto(atomic_pattern, 100) ->
 106;
yeccgoto(atomic_pattern, 114) ->
 106;
yeccgoto(atomic_pattern, 115) ->
 106;
yeccgoto(atomic_pattern, 120) ->
 106;
yeccgoto(atomic_pattern, 162) ->
 106;
yeccgoto(atomic_pattern, 174) ->
 106;
yeccgoto(atomic_pattern, 177) ->
 106;
yeccgoto(atomic_pattern, 197) ->
 106;
yeccgoto(atomic_pattern, 199) ->
 106;
yeccgoto(atomic_pattern, 229) ->
 106;
yeccgoto(attribute, 282) ->
 285;
yeccgoto(attribute, 288) ->
 285;
yeccgoto(attribute_list, 282) ->
 286;
yeccgoto(attribute_list, 288) ->
 289;
yeccgoto(binary, 33) ->
 43;
yeccgoto(binary, 34) ->
 43;
yeccgoto(binary, 35) ->
 43;
yeccgoto(binary, 36) ->
 43;
yeccgoto(binary, 37) ->
 43;
yeccgoto(binary, 40) ->
 43;
yeccgoto(binary, 44) ->
 43;
yeccgoto(binary, 46) ->
 43;
yeccgoto(binary, 48) ->
 43;
yeccgoto(binary, 52) ->
 43;
yeccgoto(binary, 70) ->
 43;
yeccgoto(binary, 74) ->
 43;
yeccgoto(binary, 79) ->
 43;
yeccgoto(binary, 86) ->
 43;
yeccgoto(binary, 90) ->
 43;
yeccgoto(binary, 99) ->
 43;
yeccgoto(binary, 164) ->
 43;
yeccgoto(binary, 166) ->
 43;
yeccgoto(binary, 171) ->
 43;
yeccgoto(binary, 212) ->
 43;
yeccgoto(binary, 218) ->
 43;
yeccgoto(binary, 221) ->
 43;
yeccgoto(binary, 223) ->
 43;
yeccgoto(binary, 225) ->
 43;
yeccgoto(binary, 233) ->
 43;
yeccgoto(binary, 239) ->
 43;
yeccgoto(binary, 242) ->
 43;
yeccgoto(binary, 258) ->
 43;
yeccgoto(binary, 267) ->
 43;
yeccgoto(binary, 271) ->
 43;
yeccgoto(binary_pattern, 65) ->
 107;
yeccgoto(binary_pattern, 95) ->
 107;
yeccgoto(binary_pattern, 96) ->
 107;
yeccgoto(binary_pattern, 97) ->
 107;
yeccgoto(binary_pattern, 98) ->
 107;
yeccgoto(binary_pattern, 100) ->
 107;
yeccgoto(binary_pattern, 114) ->
 107;
yeccgoto(binary_pattern, 115) ->
 107;
yeccgoto(binary_pattern, 120) ->
 107;
yeccgoto(binary_pattern, 162) ->
 107;
yeccgoto(binary_pattern, 174) ->
 107;
yeccgoto(binary_pattern, 177) ->
 107;
yeccgoto(binary_pattern, 197) ->
 107;
yeccgoto(binary_pattern, 199) ->
 107;
yeccgoto(binary_pattern, 229) ->
 107;
yeccgoto(call_expr, 33) ->
 45;
yeccgoto(call_expr, 34) ->
 45;
yeccgoto(call_expr, 35) ->
 45;
yeccgoto(call_expr, 36) ->
 45;
yeccgoto(call_expr, 37) ->
 45;
yeccgoto(call_expr, 40) ->
 45;
yeccgoto(call_expr, 44) ->
 45;
yeccgoto(call_expr, 46) ->
 45;
yeccgoto(call_expr, 48) ->
 45;
yeccgoto(call_expr, 52) ->
 45;
yeccgoto(call_expr, 70) ->
 45;
yeccgoto(call_expr, 74) ->
 45;
yeccgoto(call_expr, 79) ->
 45;
yeccgoto(call_expr, 86) ->
 45;
yeccgoto(call_expr, 90) ->
 45;
yeccgoto(call_expr, 99) ->
 45;
yeccgoto(call_expr, 164) ->
 45;
yeccgoto(call_expr, 166) ->
 45;
yeccgoto(call_expr, 171) ->
 45;
yeccgoto(call_expr, 212) ->
 45;
yeccgoto(call_expr, 218) ->
 45;
yeccgoto(call_expr, 221) ->
 45;
yeccgoto(call_expr, 223) ->
 45;
yeccgoto(call_expr, 225) ->
 45;
yeccgoto(call_expr, 233) ->
 45;
yeccgoto(call_expr, 239) ->
 45;
yeccgoto(call_expr, 242) ->
 45;
yeccgoto(call_expr, 258) ->
 45;
yeccgoto(call_expr, 267) ->
 45;
yeccgoto(call_expr, 271) ->
 45;
yeccgoto(case_expr, 33) ->
 47;
yeccgoto(case_expr, 34) ->
 47;
yeccgoto(case_expr, 35) ->
 47;
yeccgoto(case_expr, 36) ->
 47;
yeccgoto(case_expr, 37) ->
 47;
yeccgoto(case_expr, 40) ->
 47;
yeccgoto(case_expr, 44) ->
 47;
yeccgoto(case_expr, 46) ->
 47;
yeccgoto(case_expr, 48) ->
 47;
yeccgoto(case_expr, 52) ->
 47;
yeccgoto(case_expr, 70) ->
 47;
yeccgoto(case_expr, 74) ->
 47;
yeccgoto(case_expr, 79) ->
 47;
yeccgoto(case_expr, 86) ->
 47;
yeccgoto(case_expr, 90) ->
 47;
yeccgoto(case_expr, 99) ->
 47;
yeccgoto(case_expr, 164) ->
 47;
yeccgoto(case_expr, 166) ->
 47;
yeccgoto(case_expr, 171) ->
 47;
yeccgoto(case_expr, 212) ->
 47;
yeccgoto(case_expr, 218) ->
 47;
yeccgoto(case_expr, 221) ->
 47;
yeccgoto(case_expr, 223) ->
 47;
yeccgoto(case_expr, 225) ->
 47;
yeccgoto(case_expr, 233) ->
 47;
yeccgoto(case_expr, 239) ->
 47;
yeccgoto(case_expr, 242) ->
 47;
yeccgoto(case_expr, 258) ->
 47;
yeccgoto(case_expr, 267) ->
 47;
yeccgoto(case_expr, 271) ->
 47;
yeccgoto(catch_expr, 33) ->
 49;
yeccgoto(catch_expr, 34) ->
 49;
yeccgoto(catch_expr, 35) ->
 49;
yeccgoto(catch_expr, 36) ->
 49;
yeccgoto(catch_expr, 37) ->
 49;
yeccgoto(catch_expr, 40) ->
 49;
yeccgoto(catch_expr, 44) ->
 49;
yeccgoto(catch_expr, 46) ->
 49;
yeccgoto(catch_expr, 48) ->
 49;
yeccgoto(catch_expr, 52) ->
 49;
yeccgoto(catch_expr, 70) ->
 49;
yeccgoto(catch_expr, 74) ->
 49;
yeccgoto(catch_expr, 79) ->
 49;
yeccgoto(catch_expr, 86) ->
 49;
yeccgoto(catch_expr, 90) ->
 49;
yeccgoto(catch_expr, 99) ->
 49;
yeccgoto(catch_expr, 164) ->
 49;
yeccgoto(catch_expr, 166) ->
 49;
yeccgoto(catch_expr, 171) ->
 49;
yeccgoto(catch_expr, 212) ->
 49;
yeccgoto(catch_expr, 218) ->
 49;
yeccgoto(catch_expr, 221) ->
 49;
yeccgoto(catch_expr, 223) ->
 49;
yeccgoto(catch_expr, 225) ->
 49;
yeccgoto(catch_expr, 233) ->
 49;
yeccgoto(catch_expr, 239) ->
 49;
yeccgoto(catch_expr, 242) ->
 49;
yeccgoto(catch_expr, 258) ->
 49;
yeccgoto(catch_expr, 267) ->
 49;
yeccgoto(catch_expr, 271) ->
 49;
yeccgoto(clause, 65) ->
 108;
yeccgoto(clause, 96) ->
 185;
yeccgoto(clause, 100) ->
 108;
yeccgoto(clause, 229) ->
 108;
yeccgoto(clause_pattern, 65) ->
 109;
yeccgoto(clause_pattern, 96) ->
 109;
yeccgoto(clause_pattern, 100) ->
 109;
yeccgoto(clause_pattern, 229) ->
 109;
yeccgoto(cons, 33) ->
 51;
yeccgoto(cons, 34) ->
 51;
yeccgoto(cons, 35) ->
 51;
yeccgoto(cons, 36) ->
 51;
yeccgoto(cons, 37) ->
 51;
yeccgoto(cons, 40) ->
 51;
yeccgoto(cons, 44) ->
 51;
yeccgoto(cons, 46) ->
 51;
yeccgoto(cons, 48) ->
 51;
yeccgoto(cons, 52) ->
 51;
yeccgoto(cons, 70) ->
 51;
yeccgoto(cons, 74) ->
 51;
yeccgoto(cons, 79) ->
 51;
yeccgoto(cons, 86) ->
 51;
yeccgoto(cons, 90) ->
 51;
yeccgoto(cons, 99) ->
 51;
yeccgoto(cons, 164) ->
 51;
yeccgoto(cons, 166) ->
 51;
yeccgoto(cons, 171) ->
 51;
yeccgoto(cons, 212) ->
 51;
yeccgoto(cons, 218) ->
 51;
yeccgoto(cons, 221) ->
 51;
yeccgoto(cons, 223) ->
 51;
yeccgoto(cons, 225) ->
 51;
yeccgoto(cons, 233) ->
 51;
yeccgoto(cons, 239) ->
 51;
yeccgoto(cons, 242) ->
 51;
yeccgoto(cons, 258) ->
 51;
yeccgoto(cons, 267) ->
 51;
yeccgoto(cons, 271) ->
 51;
yeccgoto(cons_constant, 126) ->
 134;
yeccgoto(cons_constant, 129) ->
 134;
yeccgoto(cons_constant, 142) ->
 134;
yeccgoto(cons_constant, 147) ->
 134;
yeccgoto(cons_constant, 151) ->
 134;
yeccgoto(cons_constant, 154) ->
 134;
yeccgoto(cons_literal, 290) ->
 293;
yeccgoto(cons_literal, 291) ->
 293;
yeccgoto(cons_literal, 296) ->
 293;
yeccgoto(cons_literal, 301) ->
 293;
yeccgoto(cons_literal, 304) ->
 293;
yeccgoto(cons_literal, 307) ->
 293;
yeccgoto(cons_pattern, 65) ->
 110;
yeccgoto(cons_pattern, 95) ->
 110;
yeccgoto(cons_pattern, 96) ->
 110;
yeccgoto(cons_pattern, 97) ->
 110;
yeccgoto(cons_pattern, 98) ->
 110;
yeccgoto(cons_pattern, 100) ->
 110;
yeccgoto(cons_pattern, 114) ->
 110;
yeccgoto(cons_pattern, 115) ->
 110;
yeccgoto(cons_pattern, 120) ->
 110;
yeccgoto(cons_pattern, 162) ->
 110;
yeccgoto(cons_pattern, 174) ->
 110;
yeccgoto(cons_pattern, 177) ->
 110;
yeccgoto(cons_pattern, 197) ->
 110;
yeccgoto(cons_pattern, 199) ->
 110;
yeccgoto(cons_pattern, 229) ->
 110;
yeccgoto(constant, 126) ->
 135;
yeccgoto(constant, 129) ->
 150;
yeccgoto(constant, 142) ->
 135;
yeccgoto(constant, 147) ->
 135;
yeccgoto(constant, 151) ->
 157;
yeccgoto(constant, 154) ->
 155;
yeccgoto(constants, 126) ->
 136;
yeccgoto(constants, 142) ->
 143;
yeccgoto(constants, 147) ->
 148;
yeccgoto(exported_name, 5) ->
 313;
yeccgoto(exported_name, 317) ->
 313;
yeccgoto(exported_names, 5) ->
 314;
yeccgoto(exported_names, 317) ->
 318;
yeccgoto(expression, 33) ->
 53;
yeccgoto(expression, 35) ->
 250;
yeccgoto(expression, 36) ->
 53;
yeccgoto(expression, 37) ->
 53;
yeccgoto(expression, 40) ->
 53;
yeccgoto(expression, 44) ->
 53;
yeccgoto(expression, 46) ->
 53;
yeccgoto(expression, 48) ->
 53;
yeccgoto(expression, 52) ->
 53;
yeccgoto(expression, 70) ->
 53;
yeccgoto(expression, 74) ->
 53;
yeccgoto(expression, 79) ->
 53;
yeccgoto(expression, 86) ->
 53;
yeccgoto(expression, 90) ->
 53;
yeccgoto(expression, 99) ->
 53;
yeccgoto(expression, 164) ->
 53;
yeccgoto(expression, 166) ->
 53;
yeccgoto(expression, 171) ->
 53;
yeccgoto(expression, 212) ->
 53;
yeccgoto(expression, 218) ->
 53;
yeccgoto(expression, 221) ->
 53;
yeccgoto(expression, 223) ->
 53;
yeccgoto(expression, 225) ->
 53;
yeccgoto(expression, 233) ->
 53;
yeccgoto(expression, 239) ->
 53;
yeccgoto(expression, 242) ->
 53;
yeccgoto(expression, 271) ->
 53;
yeccgoto(fun_expr, 20) ->
 24;
yeccgoto(fun_expr, 21) ->
 274;
yeccgoto(fun_expr, 33) ->
 55;
yeccgoto(fun_expr, 34) ->
 55;
yeccgoto(fun_expr, 35) ->
 55;
yeccgoto(fun_expr, 36) ->
 55;
yeccgoto(fun_expr, 37) ->
 55;
yeccgoto(fun_expr, 40) ->
 55;
yeccgoto(fun_expr, 44) ->
 55;
yeccgoto(fun_expr, 46) ->
 55;
yeccgoto(fun_expr, 48) ->
 55;
yeccgoto(fun_expr, 52) ->
 55;
yeccgoto(fun_expr, 70) ->
 55;
yeccgoto(fun_expr, 74) ->
 55;
yeccgoto(fun_expr, 79) ->
 55;
yeccgoto(fun_expr, 86) ->
 55;
yeccgoto(fun_expr, 90) ->
 55;
yeccgoto(fun_expr, 99) ->
 55;
yeccgoto(fun_expr, 164) ->
 55;
yeccgoto(fun_expr, 166) ->
 55;
yeccgoto(fun_expr, 171) ->
 55;
yeccgoto(fun_expr, 212) ->
 55;
yeccgoto(fun_expr, 218) ->
 55;
yeccgoto(fun_expr, 221) ->
 55;
yeccgoto(fun_expr, 223) ->
 55;
yeccgoto(fun_expr, 225) ->
 55;
yeccgoto(fun_expr, 233) ->
 55;
yeccgoto(fun_expr, 239) ->
 55;
yeccgoto(fun_expr, 242) ->
 55;
yeccgoto(fun_expr, 258) ->
 55;
yeccgoto(fun_expr, 267) ->
 55;
yeccgoto(fun_expr, 271) ->
 55;
yeccgoto(function_definition, 8) ->
 12;
yeccgoto(function_definition, 12) ->
 12;
yeccgoto(function_definition, 60) ->
 12;
yeccgoto(function_definition, 322) ->
 12;
yeccgoto(function_definitions, 8) ->
 13;
yeccgoto(function_definitions, 12) ->
 17;
yeccgoto(function_definitions, 60) ->
 217;
yeccgoto(function_definitions, 322) ->
 13;
yeccgoto(function_name, 5) ->
 315;
yeccgoto(function_name, 8) ->
 14;
yeccgoto(function_name, 9) ->
 278;
yeccgoto(function_name, 12) ->
 14;
yeccgoto(function_name, 33) ->
 56;
yeccgoto(function_name, 34) ->
 56;
yeccgoto(function_name, 35) ->
 56;
yeccgoto(function_name, 36) ->
 56;
yeccgoto(function_name, 37) ->
 56;
yeccgoto(function_name, 40) ->
 56;
yeccgoto(function_name, 44) ->
 56;
yeccgoto(function_name, 46) ->
 56;
yeccgoto(function_name, 48) ->
 56;
yeccgoto(function_name, 52) ->
 56;
yeccgoto(function_name, 60) ->
 14;
yeccgoto(function_name, 70) ->
 56;
yeccgoto(function_name, 74) ->
 56;
yeccgoto(function_name, 79) ->
 56;
yeccgoto(function_name, 86) ->
 56;
yeccgoto(function_name, 90) ->
 56;
yeccgoto(function_name, 99) ->
 56;
yeccgoto(function_name, 164) ->
 56;
yeccgoto(function_name, 166) ->
 56;
yeccgoto(function_name, 171) ->
 56;
yeccgoto(function_name, 212) ->
 56;
yeccgoto(function_name, 218) ->
 56;
yeccgoto(function_name, 221) ->
 56;
yeccgoto(function_name, 223) ->
 56;
yeccgoto(function_name, 225) ->
 56;
yeccgoto(function_name, 233) ->
 56;
yeccgoto(function_name, 239) ->
 56;
yeccgoto(function_name, 242) ->
 56;
yeccgoto(function_name, 258) ->
 56;
yeccgoto(function_name, 267) ->
 56;
yeccgoto(function_name, 271) ->
 56;
yeccgoto(function_name, 317) ->
 315;
yeccgoto(function_name, 322) ->
 14;
yeccgoto(let_expr, 33) ->
 59;
yeccgoto(let_expr, 34) ->
 59;
yeccgoto(let_expr, 35) ->
 59;
yeccgoto(let_expr, 36) ->
 59;
yeccgoto(let_expr, 37) ->
 59;
yeccgoto(let_expr, 40) ->
 59;
yeccgoto(let_expr, 44) ->
 59;
yeccgoto(let_expr, 46) ->
 59;
yeccgoto(let_expr, 48) ->
 59;
yeccgoto(let_expr, 52) ->
 59;
yeccgoto(let_expr, 70) ->
 59;
yeccgoto(let_expr, 74) ->
 59;
yeccgoto(let_expr, 79) ->
 59;
yeccgoto(let_expr, 86) ->
 59;
yeccgoto(let_expr, 90) ->
 59;
yeccgoto(let_expr, 99) ->
 59;
yeccgoto(let_expr, 164) ->
 59;
yeccgoto(let_expr, 166) ->
 59;
yeccgoto(let_expr, 171) ->
 59;
yeccgoto(let_expr, 212) ->
 59;
yeccgoto(let_expr, 218) ->
 59;
yeccgoto(let_expr, 221) ->
 59;
yeccgoto(let_expr, 223) ->
 59;
yeccgoto(let_expr, 225) ->
 59;
yeccgoto(let_expr, 233) ->
 59;
yeccgoto(let_expr, 239) ->
 59;
yeccgoto(let_expr, 242) ->
 59;
yeccgoto(let_expr, 258) ->
 59;
yeccgoto(let_expr, 267) ->
 59;
yeccgoto(let_expr, 271) ->
 59;
yeccgoto(let_vars, 58) ->
 220;
yeccgoto(let_vars, 82) ->
 85;
yeccgoto(let_vars, 88) ->
 89;
yeccgoto(letrec_expr, 33) ->
 61;
yeccgoto(letrec_expr, 34) ->
 61;
yeccgoto(letrec_expr, 35) ->
 61;
yeccgoto(letrec_expr, 36) ->
 61;
yeccgoto(letrec_expr, 37) ->
 61;
yeccgoto(letrec_expr, 40) ->
 61;
yeccgoto(letrec_expr, 44) ->
 61;
yeccgoto(letrec_expr, 46) ->
 61;
yeccgoto(letrec_expr, 48) ->
 61;
yeccgoto(letrec_expr, 52) ->
 61;
yeccgoto(letrec_expr, 70) ->
 61;
yeccgoto(letrec_expr, 74) ->
 61;
yeccgoto(letrec_expr, 79) ->
 61;
yeccgoto(letrec_expr, 86) ->
 61;
yeccgoto(letrec_expr, 90) ->
 61;
yeccgoto(letrec_expr, 99) ->
 61;
yeccgoto(letrec_expr, 164) ->
 61;
yeccgoto(letrec_expr, 166) ->
 61;
yeccgoto(letrec_expr, 171) ->
 61;
yeccgoto(letrec_expr, 212) ->
 61;
yeccgoto(letrec_expr, 218) ->
 61;
yeccgoto(letrec_expr, 221) ->
 61;
yeccgoto(letrec_expr, 223) ->
 61;
yeccgoto(letrec_expr, 225) ->
 61;
yeccgoto(letrec_expr, 233) ->
 61;
yeccgoto(letrec_expr, 239) ->
 61;
yeccgoto(letrec_expr, 242) ->
 61;
yeccgoto(letrec_expr, 258) ->
 61;
yeccgoto(letrec_expr, 267) ->
 61;
yeccgoto(letrec_expr, 271) ->
 61;
yeccgoto(literal, 290) ->
 294;
yeccgoto(literal, 291) ->
 303;
yeccgoto(literal, 296) ->
 297;
yeccgoto(literal, 301) ->
 297;
yeccgoto(literal, 304) ->
 310;
yeccgoto(literal, 307) ->
 308;
yeccgoto(literals, 296) ->
 298;
yeccgoto(literals, 301) ->
 302;
yeccgoto(module_attribute, 6) ->
 8;
yeccgoto(module_attribute, 321) ->
 322;
yeccgoto(module_definition, 0) ->
 3;
yeccgoto(module_defs, 8) ->
 15;
yeccgoto(module_defs, 322) ->
 323;
yeccgoto(module_export, 4) ->
 6;
yeccgoto(module_export, 320) ->
 321;
yeccgoto(nil, 33) ->
 62;
yeccgoto(nil, 34) ->
 62;
yeccgoto(nil, 35) ->
 62;
yeccgoto(nil, 36) ->
 62;
yeccgoto(nil, 37) ->
 62;
yeccgoto(nil, 40) ->
 62;
yeccgoto(nil, 44) ->
 62;
yeccgoto(nil, 46) ->
 62;
yeccgoto(nil, 48) ->
 62;
yeccgoto(nil, 52) ->
 62;
yeccgoto(nil, 65) ->
 62;
yeccgoto(nil, 70) ->
 62;
yeccgoto(nil, 74) ->
 62;
yeccgoto(nil, 79) ->
 62;
yeccgoto(nil, 86) ->
 62;
yeccgoto(nil, 90) ->
 62;
yeccgoto(nil, 95) ->
 62;
yeccgoto(nil, 96) ->
 62;
yeccgoto(nil, 97) ->
 62;
yeccgoto(nil, 98) ->
 62;
yeccgoto(nil, 99) ->
 62;
yeccgoto(nil, 100) ->
 62;
yeccgoto(nil, 114) ->
 62;
yeccgoto(nil, 115) ->
 62;
yeccgoto(nil, 120) ->
 62;
yeccgoto(nil, 126) ->
 139;
yeccgoto(nil, 129) ->
 139;
yeccgoto(nil, 142) ->
 139;
yeccgoto(nil, 147) ->
 139;
yeccgoto(nil, 151) ->
 139;
yeccgoto(nil, 154) ->
 139;
yeccgoto(nil, 162) ->
 62;
yeccgoto(nil, 164) ->
 62;
yeccgoto(nil, 166) ->
 62;
yeccgoto(nil, 171) ->
 62;
yeccgoto(nil, 174) ->
 62;
yeccgoto(nil, 177) ->
 62;
yeccgoto(nil, 197) ->
 62;
yeccgoto(nil, 199) ->
 62;
yeccgoto(nil, 212) ->
 62;
yeccgoto(nil, 218) ->
 62;
yeccgoto(nil, 221) ->
 62;
yeccgoto(nil, 223) ->
 62;
yeccgoto(nil, 225) ->
 62;
yeccgoto(nil, 229) ->
 62;
yeccgoto(nil, 233) ->
 62;
yeccgoto(nil, 239) ->
 62;
yeccgoto(nil, 242) ->
 62;
yeccgoto(nil, 258) ->
 62;
yeccgoto(nil, 267) ->
 62;
yeccgoto(nil, 271) ->
 62;
yeccgoto(nil, 290) ->
 62;
yeccgoto(nil, 291) ->
 62;
yeccgoto(nil, 296) ->
 62;
yeccgoto(nil, 301) ->
 62;
yeccgoto(nil, 304) ->
 62;
yeccgoto(nil, 307) ->
 62;
yeccgoto(other_pattern, 65) ->
 111;
yeccgoto(other_pattern, 95) ->
 191;
yeccgoto(other_pattern, 96) ->
 186;
yeccgoto(other_pattern, 97) ->
 111;
yeccgoto(other_pattern, 98) ->
 111;
yeccgoto(other_pattern, 100) ->
 111;
yeccgoto(other_pattern, 114) ->
 111;
yeccgoto(other_pattern, 115) ->
 123;
yeccgoto(other_pattern, 120) ->
 111;
yeccgoto(other_pattern, 162) ->
 111;
yeccgoto(other_pattern, 174) ->
 111;
yeccgoto(other_pattern, 177) ->
 111;
yeccgoto(other_pattern, 197) ->
 191;
yeccgoto(other_pattern, 199) ->
 191;
yeccgoto(other_pattern, 229) ->
 111;
yeccgoto(pattern, 95) ->
 192;
yeccgoto(pattern, 197) ->
 192;
yeccgoto(pattern, 199) ->
 200;
yeccgoto(primop_expr, 33) ->
 64;
yeccgoto(primop_expr, 34) ->
 64;
yeccgoto(primop_expr, 35) ->
 64;
yeccgoto(primop_expr, 36) ->
 64;
yeccgoto(primop_expr, 37) ->
 64;
yeccgoto(primop_expr, 40) ->
 64;
yeccgoto(primop_expr, 44) ->
 64;
yeccgoto(primop_expr, 46) ->
 64;
yeccgoto(primop_expr, 48) ->
 64;
yeccgoto(primop_expr, 52) ->
 64;
yeccgoto(primop_expr, 70) ->
 64;
yeccgoto(primop_expr, 74) ->
 64;
yeccgoto(primop_expr, 79) ->
 64;
yeccgoto(primop_expr, 86) ->
 64;
yeccgoto(primop_expr, 90) ->
 64;
yeccgoto(primop_expr, 99) ->
 64;
yeccgoto(primop_expr, 164) ->
 64;
yeccgoto(primop_expr, 166) ->
 64;
yeccgoto(primop_expr, 171) ->
 64;
yeccgoto(primop_expr, 212) ->
 64;
yeccgoto(primop_expr, 218) ->
 64;
yeccgoto(primop_expr, 221) ->
 64;
yeccgoto(primop_expr, 223) ->
 64;
yeccgoto(primop_expr, 225) ->
 64;
yeccgoto(primop_expr, 233) ->
 64;
yeccgoto(primop_expr, 239) ->
 64;
yeccgoto(primop_expr, 242) ->
 64;
yeccgoto(primop_expr, 258) ->
 64;
yeccgoto(primop_expr, 267) ->
 64;
yeccgoto(primop_expr, 271) ->
 64;
yeccgoto(receive_expr, 33) ->
 66;
yeccgoto(receive_expr, 34) ->
 66;
yeccgoto(receive_expr, 35) ->
 66;
yeccgoto(receive_expr, 36) ->
 66;
yeccgoto(receive_expr, 37) ->
 66;
yeccgoto(receive_expr, 40) ->
 66;
yeccgoto(receive_expr, 44) ->
 66;
yeccgoto(receive_expr, 46) ->
 66;
yeccgoto(receive_expr, 48) ->
 66;
yeccgoto(receive_expr, 52) ->
 66;
yeccgoto(receive_expr, 70) ->
 66;
yeccgoto(receive_expr, 74) ->
 66;
yeccgoto(receive_expr, 79) ->
 66;
yeccgoto(receive_expr, 86) ->
 66;
yeccgoto(receive_expr, 90) ->
 66;
yeccgoto(receive_expr, 99) ->
 66;
yeccgoto(receive_expr, 164) ->
 66;
yeccgoto(receive_expr, 166) ->
 66;
yeccgoto(receive_expr, 171) ->
 66;
yeccgoto(receive_expr, 212) ->
 66;
yeccgoto(receive_expr, 218) ->
 66;
yeccgoto(receive_expr, 221) ->
 66;
yeccgoto(receive_expr, 223) ->
 66;
yeccgoto(receive_expr, 225) ->
 66;
yeccgoto(receive_expr, 233) ->
 66;
yeccgoto(receive_expr, 239) ->
 66;
yeccgoto(receive_expr, 242) ->
 66;
yeccgoto(receive_expr, 258) ->
 66;
yeccgoto(receive_expr, 267) ->
 66;
yeccgoto(receive_expr, 271) ->
 66;
yeccgoto(seg_flags, 204) ->
 206;
yeccgoto(seg_flags, 207) ->
 208;
yeccgoto(seg_flags, 264) ->
 265;
yeccgoto(seg_size_pattern, 199) ->
 201;
yeccgoto(seg_size_unit, 258) ->
 259;
yeccgoto(seg_type, 202) ->
 204;
yeccgoto(seg_type, 263) ->
 264;
yeccgoto(segment, 34) ->
 255;
yeccgoto(segment, 267) ->
 255;
yeccgoto(segment_pattern, 95) ->
 193;
yeccgoto(segment_pattern, 197) ->
 193;
yeccgoto(segment_patterns, 95) ->
 194;
yeccgoto(segment_patterns, 197) ->
 198;
yeccgoto(segments, 34) ->
 256;
yeccgoto(segments, 267) ->
 268;
yeccgoto(sequence, 33) ->
 67;
yeccgoto(sequence, 34) ->
 67;
yeccgoto(sequence, 35) ->
 67;
yeccgoto(sequence, 36) ->
 67;
yeccgoto(sequence, 37) ->
 67;
yeccgoto(sequence, 40) ->
 67;
yeccgoto(sequence, 44) ->
 67;
yeccgoto(sequence, 46) ->
 67;
yeccgoto(sequence, 48) ->
 67;
yeccgoto(sequence, 52) ->
 67;
yeccgoto(sequence, 70) ->
 67;
yeccgoto(sequence, 74) ->
 67;
yeccgoto(sequence, 79) ->
 67;
yeccgoto(sequence, 86) ->
 67;
yeccgoto(sequence, 90) ->
 67;
yeccgoto(sequence, 99) ->
 67;
yeccgoto(sequence, 164) ->
 67;
yeccgoto(sequence, 166) ->
 67;
yeccgoto(sequence, 171) ->
 67;
yeccgoto(sequence, 212) ->
 67;
yeccgoto(sequence, 218) ->
 67;
yeccgoto(sequence, 221) ->
 67;
yeccgoto(sequence, 223) ->
 67;
yeccgoto(sequence, 225) ->
 67;
yeccgoto(sequence, 233) ->
 67;
yeccgoto(sequence, 239) ->
 67;
yeccgoto(sequence, 242) ->
 67;
yeccgoto(sequence, 258) ->
 67;
yeccgoto(sequence, 267) ->
 67;
yeccgoto(sequence, 271) ->
 67;
yeccgoto(single_expression, 33) ->
 68;
yeccgoto(single_expression, 34) ->
 257;
yeccgoto(single_expression, 35) ->
 68;
yeccgoto(single_expression, 36) ->
 68;
yeccgoto(single_expression, 37) ->
 68;
yeccgoto(single_expression, 40) ->
 68;
yeccgoto(single_expression, 44) ->
 68;
yeccgoto(single_expression, 46) ->
 68;
yeccgoto(single_expression, 48) ->
 68;
yeccgoto(single_expression, 52) ->
 68;
yeccgoto(single_expression, 70) ->
 68;
yeccgoto(single_expression, 74) ->
 68;
yeccgoto(single_expression, 79) ->
 68;
yeccgoto(single_expression, 86) ->
 68;
yeccgoto(single_expression, 90) ->
 68;
yeccgoto(single_expression, 99) ->
 68;
yeccgoto(single_expression, 164) ->
 68;
yeccgoto(single_expression, 166) ->
 68;
yeccgoto(single_expression, 171) ->
 68;
yeccgoto(single_expression, 212) ->
 68;
yeccgoto(single_expression, 218) ->
 68;
yeccgoto(single_expression, 221) ->
 68;
yeccgoto(single_expression, 223) ->
 68;
yeccgoto(single_expression, 225) ->
 68;
yeccgoto(single_expression, 233) ->
 68;
yeccgoto(single_expression, 239) ->
 68;
yeccgoto(single_expression, 242) ->
 68;
yeccgoto(single_expression, 258) ->
 260;
yeccgoto(single_expression, 267) ->
 257;
yeccgoto(single_expression, 271) ->
 68;
yeccgoto(tail, 238) ->
 241;
yeccgoto(tail, 245) ->
 246;
yeccgoto(tail_constant, 150) ->
 153;
yeccgoto(tail_constant, 157) ->
 158;
yeccgoto(tail_literal, 303) ->
 306;
yeccgoto(tail_literal, 310) ->
 311;
yeccgoto(tail_pattern, 173) ->
 176;
yeccgoto(tail_pattern, 180) ->
 181;
yeccgoto(timeout, 65) ->
 112;
yeccgoto(timeout, 101) ->
 168;
yeccgoto(try_expr, 33) ->
 71;
yeccgoto(try_expr, 34) ->
 71;
yeccgoto(try_expr, 35) ->
 71;
yeccgoto(try_expr, 36) ->
 71;
yeccgoto(try_expr, 37) ->
 71;
yeccgoto(try_expr, 40) ->
 71;
yeccgoto(try_expr, 44) ->
 71;
yeccgoto(try_expr, 46) ->
 71;
yeccgoto(try_expr, 48) ->
 71;
yeccgoto(try_expr, 52) ->
 71;
yeccgoto(try_expr, 70) ->
 71;
yeccgoto(try_expr, 74) ->
 71;
yeccgoto(try_expr, 79) ->
 71;
yeccgoto(try_expr, 86) ->
 71;
yeccgoto(try_expr, 90) ->
 71;
yeccgoto(try_expr, 99) ->
 71;
yeccgoto(try_expr, 164) ->
 71;
yeccgoto(try_expr, 166) ->
 71;
yeccgoto(try_expr, 171) ->
 71;
yeccgoto(try_expr, 212) ->
 71;
yeccgoto(try_expr, 218) ->
 71;
yeccgoto(try_expr, 221) ->
 71;
yeccgoto(try_expr, 223) ->
 71;
yeccgoto(try_expr, 225) ->
 71;
yeccgoto(try_expr, 233) ->
 71;
yeccgoto(try_expr, 239) ->
 71;
yeccgoto(try_expr, 242) ->
 71;
yeccgoto(try_expr, 258) ->
 71;
yeccgoto(try_expr, 267) ->
 71;
yeccgoto(try_expr, 271) ->
 71;
yeccgoto(tuple, 33) ->
 72;
yeccgoto(tuple, 34) ->
 72;
yeccgoto(tuple, 35) ->
 72;
yeccgoto(tuple, 36) ->
 72;
yeccgoto(tuple, 37) ->
 72;
yeccgoto(tuple, 40) ->
 72;
yeccgoto(tuple, 44) ->
 72;
yeccgoto(tuple, 46) ->
 72;
yeccgoto(tuple, 48) ->
 72;
yeccgoto(tuple, 52) ->
 72;
yeccgoto(tuple, 70) ->
 72;
yeccgoto(tuple, 74) ->
 72;
yeccgoto(tuple, 79) ->
 72;
yeccgoto(tuple, 86) ->
 72;
yeccgoto(tuple, 90) ->
 72;
yeccgoto(tuple, 99) ->
 72;
yeccgoto(tuple, 164) ->
 72;
yeccgoto(tuple, 166) ->
 72;
yeccgoto(tuple, 171) ->
 72;
yeccgoto(tuple, 212) ->
 72;
yeccgoto(tuple, 218) ->
 72;
yeccgoto(tuple, 221) ->
 72;
yeccgoto(tuple, 223) ->
 72;
yeccgoto(tuple, 225) ->
 72;
yeccgoto(tuple, 233) ->
 72;
yeccgoto(tuple, 239) ->
 72;
yeccgoto(tuple, 242) ->
 72;
yeccgoto(tuple, 258) ->
 72;
yeccgoto(tuple, 267) ->
 72;
yeccgoto(tuple, 271) ->
 72;
yeccgoto(tuple_constant, 126) ->
 141;
yeccgoto(tuple_constant, 129) ->
 141;
yeccgoto(tuple_constant, 142) ->
 141;
yeccgoto(tuple_constant, 147) ->
 141;
yeccgoto(tuple_constant, 151) ->
 141;
yeccgoto(tuple_constant, 154) ->
 141;
yeccgoto(tuple_literal, 290) ->
 295;
yeccgoto(tuple_literal, 291) ->
 295;
yeccgoto(tuple_literal, 296) ->
 295;
yeccgoto(tuple_literal, 301) ->
 295;
yeccgoto(tuple_literal, 304) ->
 295;
yeccgoto(tuple_literal, 307) ->
 295;
yeccgoto(tuple_pattern, 65) ->
 113;
yeccgoto(tuple_pattern, 95) ->
 113;
yeccgoto(tuple_pattern, 96) ->
 113;
yeccgoto(tuple_pattern, 97) ->
 113;
yeccgoto(tuple_pattern, 98) ->
 113;
yeccgoto(tuple_pattern, 100) ->
 113;
yeccgoto(tuple_pattern, 114) ->
 113;
yeccgoto(tuple_pattern, 115) ->
 113;
yeccgoto(tuple_pattern, 120) ->
 113;
yeccgoto(tuple_pattern, 162) ->
 113;
yeccgoto(tuple_pattern, 174) ->
 113;
yeccgoto(tuple_pattern, 177) ->
 113;
yeccgoto(tuple_pattern, 197) ->
 113;
yeccgoto(tuple_pattern, 199) ->
 113;
yeccgoto(tuple_pattern, 229) ->
 113;
yeccgoto(variable, 25) ->
 31;
yeccgoto(variable, 26) ->
 273;
yeccgoto(variable, 33) ->
 73;
yeccgoto(variable, 34) ->
 73;
yeccgoto(variable, 35) ->
 73;
yeccgoto(variable, 36) ->
 73;
yeccgoto(variable, 37) ->
 73;
yeccgoto(variable, 40) ->
 73;
yeccgoto(variable, 44) ->
 73;
yeccgoto(variable, 46) ->
 73;
yeccgoto(variable, 48) ->
 73;
yeccgoto(variable, 52) ->
 73;
yeccgoto(variable, 58) ->
 31;
yeccgoto(variable, 65) ->
 31;
yeccgoto(variable, 70) ->
 73;
yeccgoto(variable, 74) ->
 73;
yeccgoto(variable, 79) ->
 73;
yeccgoto(variable, 82) ->
 31;
yeccgoto(variable, 83) ->
 31;
yeccgoto(variable, 86) ->
 73;
yeccgoto(variable, 88) ->
 31;
yeccgoto(variable, 90) ->
 73;
yeccgoto(variable, 95) ->
 195;
yeccgoto(variable, 96) ->
 124;
yeccgoto(variable, 97) ->
 31;
yeccgoto(variable, 98) ->
 31;
yeccgoto(variable, 99) ->
 73;
yeccgoto(variable, 100) ->
 31;
yeccgoto(variable, 114) ->
 31;
yeccgoto(variable, 115) ->
 124;
yeccgoto(variable, 120) ->
 31;
yeccgoto(variable, 162) ->
 31;
yeccgoto(variable, 164) ->
 73;
yeccgoto(variable, 166) ->
 73;
yeccgoto(variable, 171) ->
 73;
yeccgoto(variable, 174) ->
 31;
yeccgoto(variable, 177) ->
 31;
yeccgoto(variable, 197) ->
 195;
yeccgoto(variable, 199) ->
 195;
yeccgoto(variable, 212) ->
 73;
yeccgoto(variable, 218) ->
 73;
yeccgoto(variable, 221) ->
 73;
yeccgoto(variable, 223) ->
 73;
yeccgoto(variable, 225) ->
 73;
yeccgoto(variable, 229) ->
 31;
yeccgoto(variable, 233) ->
 73;
yeccgoto(variable, 239) ->
 73;
yeccgoto(variable, 242) ->
 73;
yeccgoto(variable, 258) ->
 73;
yeccgoto(variable, 267) ->
 73;
yeccgoto(variable, 269) ->
 31;
yeccgoto(variable, 271) ->
 73;
yeccgoto(__Symbol, __State) ->
 exit({__Symbol, __State, missing_in_goto_table}).


