-module(megaco_text_parser_v2).
-define(THIS_MODULE, megaco_text_parser_v2).
-export([parse/1, parse_and_scan/1, format_error/1]).

%% The following directive is needed for (significantly) faster compilation
%% of the generated .erl file by the HiPE compiler.  Please do not remove.
-compile([{hipe,[{regalloc,linear_scan}]}]).

-include("megaco_text_parser_v2.hrl").



%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% The parser generator will insert appropriate declarations before this line.%

parse(Tokens) ->
    case catch yeccpars1(Tokens, false, 0, [], []) of
	error ->
	    Errorline =
		if Tokens == [] -> 0; true -> element(2, hd(Tokens)) end,
	    {error,
	     {Errorline, ?THIS_MODULE, "syntax error at or after this line."}};
	Other ->
	    Other
    end.

parse_and_scan({Mod, Fun, Args}) ->
    case apply(Mod, Fun, Args) of
	{eof, _} ->
	    {ok, eof};
	{error, Descriptor, _} ->
	    {error, Descriptor};
	{ok, Tokens, _} ->
	    yeccpars1(Tokens, {Mod, Fun, Args}, 0, [], [])
    end.

format_error(Message) ->
    case io_lib:deep_char_list(Message) of
	true ->
	    Message;
	_ ->
	    io_lib:write(Message)
    end.

% To be used in grammar files to throw an error message to the parser toplevel.
% Doesn't have to be exported!
return_error(Line, Message) ->
    throw({error, {Line, ?THIS_MODULE, Message}}).


% Don't change yeccpars1/6 too much, it is called recursively by yeccpars2/8!
yeccpars1([Token | Tokens], Tokenizer, State, States, Vstack) ->
    yeccpars2(State, element(1, Token), States, Vstack, Token, Tokens,
	      Tokenizer);
yeccpars1([], {M, F, A}, State, States, Vstack) ->
    case catch apply(M, F, A) of
        {eof, Endline} ->
            {error, {Endline, ?THIS_MODULE, "end_of_file"}};
        {error, Descriptor, _Endline} ->
            {error, Descriptor};
        {'EXIT', Reason} ->
            {error, {0, ?THIS_MODULE, Reason}};
        {ok, Tokens, _Endline} ->
	    case catch yeccpars1(Tokens, {M, F, A}, State, States, Vstack) of
		error ->
		    Errorline = element(2, hd(Tokens)),
		    {error, {Errorline, ?THIS_MODULE,
			     "syntax error at or after this line."}};
		Other ->
		    Other
	    end
    end;
yeccpars1([], false, State, States, Vstack) ->
    yeccpars2(State, '$end', States, Vstack, {'$end', 999999}, [], false).

% For internal use only.
yeccerror(Token) ->
    {error,
     {element(2, Token), ?THIS_MODULE,
      ["syntax error before: ", yecctoken2string(Token)]}}.

yecctoken2string({atom, _, A}) -> io_lib:write(A);
yecctoken2string({integer,_,N}) -> io_lib:write(N);
yecctoken2string({float,_,F}) -> io_lib:write(F);
yecctoken2string({char,_,C}) -> io_lib:write_char(C);
yecctoken2string({var,_,V}) -> io_lib:format('~s', [V]);
yecctoken2string({string,_,S}) -> io_lib:write_string(S);
yecctoken2string({reserved_symbol, _, A}) -> io_lib:format('~w', [A]);
yecctoken2string({_Cat, _, Val}) -> io_lib:format('~w', [Val]);

yecctoken2string({'dot', _}) -> io_lib:format('~w', ['.']);
yecctoken2string({'$end', _}) ->
    [];
yecctoken2string({Other, _}) when atom(Other) ->
    io_lib:format('~w', [Other]);
yecctoken2string(Other) ->
    io_lib:write(Other).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


yeccpars2(0, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [0 | __Ss], [__T | __Stack]);
yeccpars2(0, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(3, __Cat, [0 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(1, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  sep,
 yeccpars2(yeccgoto(optSep, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(2, '$end', _, __Stack, _, _, _) ->
 {ok, hd(__Stack)};
yeccpars2(2, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(3, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 4, [3 | __Ss], [__T | __Stack]);
yeccpars2(3, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(5, __Cat, [3 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(4, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 853, [4 | __Ss], [__T | __Stack]);
yeccpars2(4, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(5, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [5 | __Ss], [__T | __Stack]);
yeccpars2(5, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(6, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(7, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(8, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(9, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(10, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(11, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(12, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(13, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(14, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(15, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(16, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(17, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(18, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(19, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(20, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(21, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(22, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(23, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(24, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(25, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(26, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(27, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(28, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(29, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(30, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(31, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(32, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(33, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(34, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(35, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(36, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(37, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(38, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(39, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(40, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(41, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(42, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(43, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(44, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(45, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(46, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(47, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(48, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(49, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(50, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(51, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(52, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(53, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(54, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(55, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(56, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(57, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(58, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(59, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(60, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(61, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(62, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(63, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(64, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(65, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(66, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(67, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(68, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(69, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(70, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(71, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(72, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(73, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(74, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(75, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(76, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(77, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(78, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(79, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(80, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(81, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(82, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(83, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(84, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(85, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(86, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(87, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(88, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(89, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(90, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(91, 'endOfMessage', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 852, [91 | __Ss], [__T | __Stack]);
yeccpars2(91, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(92, 'LSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'LESSER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 93, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [92 | __Ss], [__T | __Stack]);
yeccpars2(92, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(98, __Cat, [92 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(93, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [93 | __Ss], [__T | __Stack]);
yeccpars2(93, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(94, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 838, [94 | __Ss], [__T | __Stack]);
yeccpars2(94, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(839, __Cat, [94 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(95, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mId, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(96, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(mId, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(97, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 108, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 110, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 109, [97 | __Ss], [__T | __Stack]);
yeccpars2(97, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(98, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, 'MtpAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 99, [98 | __Ss], [__T | __Stack]);
yeccpars2(98, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(99, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_mtpAddress(__1),
 yeccpars2(yeccgoto(mtpAddress, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(100, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [100 | __Ss], [__T | __Stack]);
yeccpars2(100, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(105, __Cat, [100 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(101, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [101 | __Ss], [__T | __Stack]);
yeccpars2(101, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(104, __Cat, [101 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(102, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {deviceName,__1},
 yeccpars2(yeccgoto(deviceName, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(103, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_pathName(__1),
 yeccpars2(yeccgoto(pathName, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(104, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(mId, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(105, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(mId, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(106, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 822, [106 | __Ss], [__T | __Stack]);
yeccpars2(106, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(107, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 834, [107 | __Ss], [__T | __Stack]);
yeccpars2(107, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(108, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 692, [108 | __Ss], [__T | __Stack]);
yeccpars2(108, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(109, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 684, [109 | __Ss], [__T | __Stack]);
yeccpars2(109, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(110, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 120, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 121, [110 | __Ss], [__T | __Stack]);
yeccpars2(110, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(111, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {messageError,__1},
 yeccpars2(yeccgoto(messageBody, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(112, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_message(__1,__2,__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(message, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(113, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 108, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 110, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 107, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 109, [113 | __Ss], [__T | __Stack]);
yeccpars2(113, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1],
 yeccpars2(yeccgoto(transactionList, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(114, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {transactions,__1},
 yeccpars2(yeccgoto(messageBody, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(115, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {transactionPending,__1},
 yeccpars2(yeccgoto(transactionItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(116, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {transactionReply,__1},
 yeccpars2(yeccgoto(transactionItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(117, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {transactionRequest,__1},
 yeccpars2(yeccgoto(transactionItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(118, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {transactionResponseAck,__1},
 yeccpars2(yeccgoto(transactionItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(119, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(transactionList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(120, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 674, [120 | __Ss], [__T | __Stack]);
yeccpars2(120, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(121, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [121 | __Ss], [__T | __Stack]);
yeccpars2(121, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(122, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 129, [122 | __Ss], [__T | __Stack]);
yeccpars2(122, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(123, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [123 | __Ss], [__T | __Stack]);
yeccpars2(123, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(125, __Cat, [123 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(124, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [124 | __Ss], [__T | __Stack]);
yeccpars2(124, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(125, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 126, [125 | __Ss], [__T | __Stack]);
yeccpars2(125, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(126, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'TransactionRequest'{transactionId = asn1_NOVALUE, actions = [__3|__4]},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(transactionRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(127, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [127 | __Ss], [__T | __Stack]);
yeccpars2(127, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(128, __Cat, [127 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(128, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(actionRequestList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(129, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [129 | __Ss], [__T | __Stack]);
yeccpars2(129, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(130, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 132, [130 | __Ss], [__T | __Stack]);
yeccpars2(130, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(131, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_contextID(__1),
 yeccpars2(yeccgoto(contextID, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(132, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 136, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [132 | __Ss], [__T | __Stack]);
yeccpars2(132, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(133, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {addReq,__1},
 yeccpars2(yeccgoto(ammToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(134, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 671, [134 | __Ss], [__T | __Stack]);
yeccpars2(134, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(135, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 668, [135 | __Ss], [__T | __Stack]);
yeccpars2(135, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(136, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 658, [136 | __Ss], [__T | __Stack]);
yeccpars2(136, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(137, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {emergency,true},
 yeccpars2(yeccgoto(contextProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(138, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {modReq,__1},
 yeccpars2(yeccgoto(ammToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(139, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {moveReq,__1},
 yeccpars2(yeccgoto(ammToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(140, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 641, [140 | __Ss], [__T | __Stack]);
yeccpars2(140, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(141, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 639, [141 | __Ss], [__T | __Stack]);
yeccpars2(141, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(142, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 590, [142 | __Ss], [__T | __Stack]);
yeccpars2(142, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(143, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 584, [143 | __Ss], [__T | __Stack]);
yeccpars2(143, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(144, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 567, [144 | __Ss], [__T | __Stack]);
yeccpars2(144, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(145, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 562, [145 | __Ss], [__T | __Stack]);
yeccpars2(145, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(563, __Cat, [145 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(146, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(commandRequest, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(147, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 157, [147 | __Ss], [__T | __Stack]);
yeccpars2(147, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(148, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(commandRequest, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(149, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(actionRequestItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(150, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(actionRequestItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(151, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(actionRequestItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(152, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(commandRequest, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(153, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {priority,__1},
 yeccpars2(yeccgoto(contextProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(154, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(commandRequest, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(155, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(commandRequest, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(156, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {topology,__1},
 yeccpars2(yeccgoto(contextProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(157, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [157 | __Ss], [__T | __Stack]);
yeccpars2(157, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(158, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_terminationID(__1),
 yeccpars2(yeccgoto(terminationID, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(159, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 160, [159 | __Ss], [__T | __Stack]);
yeccpars2(159, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(161, __Cat, [159 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(160, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 163, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 167, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 169, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 162, [160 | __Ss], [__T | __Stack]);
yeccpars2(160, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(161, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_commandRequest(__1,#'AmmRequest'{terminationID = [__3], descriptors = __4}),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(ammRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(162, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 461, [162 | __Ss], [__T | __Stack]);
yeccpars2(162, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(163, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_DMD(__1),
 yeccpars2(yeccgoto(digitMapDescriptor, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(164, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 435, [164 | __Ss], [__T | __Stack]);
yeccpars2(164, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(yeccgoto(eventBufferDescriptor, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(165, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 371, [165 | __Ss], [__T | __Stack]);
yeccpars2(165, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'EventsDescriptor'{requestID = asn1_NOVALUE, eventList = []},
 yeccpars2(yeccgoto(eventsDescriptor, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(166, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 301, [166 | __Ss], [__T | __Stack]);
yeccpars2(166, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(167, 'LSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 281, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 280, [167 | __Ss], [__T | __Stack]);
yeccpars2(167, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(168, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 269, [168 | __Ss], [__T | __Stack]);
yeccpars2(168, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(169, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 184, [169 | __Ss], [__T | __Stack]);
yeccpars2(169, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(yeccgoto(signalsDescriptor, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(170, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 179, [170 | __Ss], [__T | __Stack]);
yeccpars2(170, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(180, __Cat, [170 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(171, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(172, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {digitMapDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(173, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {eventBufferDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(174, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {eventsDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(175, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {mediaDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(176, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {modemDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(177, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {muxDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(178, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {signalsDescriptor,__1},
 yeccpars2(yeccgoto(ammParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(179, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 163, [179 | __Ss], [__T | __Stack]);
yeccpars2(179, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 167, [179 | __Ss], [__T | __Stack]);
yeccpars2(179, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 169, [179 | __Ss], [__T | __Stack]);
yeccpars2(179, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [179 | __Ss], [__T | __Stack]);
yeccpars2(179, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 168, [179 | __Ss], [__T | __Stack]);
yeccpars2(179, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [179 | __Ss], [__T | __Stack]);
yeccpars2(179, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 166, [179 | __Ss], [__T | __Stack]);
yeccpars2(179, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 162, [179 | __Ss], [__T | __Stack]);
yeccpars2(179, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(180, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 181, [180 | __Ss], [__T | __Stack]);
yeccpars2(180, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(181, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(ammRequestBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(182, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 179, [182 | __Ss], [__T | __Stack]);
yeccpars2(182, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(183, __Cat, [182 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(183, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(ammParameters, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(184, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 185, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [184 | __Ss], [__T | __Stack]);
yeccpars2(184, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(185, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 258, [185 | __Ss], [__T | __Stack]);
yeccpars2(185, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(186, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(signalName, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(187, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_pkgdName(__1),
 yeccpars2(yeccgoto(pkgdName, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(188, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {seqSigList,__1},
 yeccpars2(yeccgoto(signalParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(189, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 197, [189 | __Ss], [__T | __Stack]);
yeccpars2(189, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_signalRequest(__1,[]),
 yeccpars2(yeccgoto(signalRequest, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(190, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 192, [190 | __Ss], [__T | __Stack]);
yeccpars2(190, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(193, __Cat, [190 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(191, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {signal,__1},
 yeccpars2(yeccgoto(signalParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(192, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 185, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [192 | __Ss], [__T | __Stack]);
yeccpars2(192, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(193, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 194, [193 | __Ss], [__T | __Stack]);
yeccpars2(193, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(194, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(signalsDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(195, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 192, [195 | __Ss], [__T | __Stack]);
yeccpars2(195, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(196, __Cat, [195 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(196, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(signalParms, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(197, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 202, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 201, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 200, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 199, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 198, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [197 | __Ss], [__T | __Stack]);
yeccpars2(197, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(198, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 256, [198 | __Ss], [__T | __Stack]);
yeccpars2(198, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(199, 'COMMA', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  keepActive,
 yeccpars2(yeccgoto(sigParameter, hd(__Ss)), 'COMMA', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(199, 'RBRKT', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  keepActive,
 yeccpars2(yeccgoto(sigParameter, hd(__Ss)), 'RBRKT', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(199, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(200, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 244, [200 | __Ss], [__T | __Stack]);
yeccpars2(200, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(201, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 239, [201 | __Ss], [__T | __Stack]);
yeccpars2(201, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(202, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 236, [202 | __Ss], [__T | __Stack]);
yeccpars2(202, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(203, 'GREATER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 211, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'LESSER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 212, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'NEQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 213, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 210, [203 | __Ss], [__T | __Stack]);
yeccpars2(203, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(204, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 205, [204 | __Ss], [__T | __Stack]);
yeccpars2(204, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(206, __Cat, [204 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(205, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 202, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 201, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 200, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 199, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 198, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [205 | __Ss], [__T | __Stack]);
yeccpars2(205, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(206, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 207, [206 | __Ss], [__T | __Stack]);
yeccpars2(206, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(207, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_signalRequest(__1,[__3|__4]),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(signalRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(208, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 205, [208 | __Ss], [__T | __Stack]);
yeccpars2(208, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(209, __Cat, [208 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(209, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(sigParameters, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(210, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 215, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'LSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 221, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 220, [210 | __Ss], [__T | __Stack]);
yeccpars2(210, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(211, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 215, [211 | __Ss], [__T | __Stack]);
yeccpars2(211, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(212, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 215, [212 | __Ss], [__T | __Stack]);
yeccpars2(212, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(213, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 215, [213 | __Ss], [__T | __Stack]);
yeccpars2(213, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(214, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {other,ensure_NAME(__1),__2},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(sigParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(215, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_value(__1),
 yeccpars2(yeccgoto(value, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(216, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_value(__1),
 yeccpars2(yeccgoto(value, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(217, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'PropertyParm'{value = [__2], extraInfo = {relation,unequalTo}},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(parmValue, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(218, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'PropertyParm'{value = [__2], extraInfo = {relation,smallerThan}},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(parmValue, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(219, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'PropertyParm'{value = [__2], extraInfo = {relation,greaterThan}},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(parmValue, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(220, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 215, [220 | __Ss], [__T | __Stack]);
yeccpars2(220, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(221, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 215, [221 | __Ss], [__T | __Stack]);
yeccpars2(221, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(222, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(parmValue, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(223, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'PropertyParm'{value = [__1]},
 yeccpars2(yeccgoto(alternativeValue, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(224, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 226, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 225, [224 | __Ss], [__T | __Stack]);
yeccpars2(224, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(227, __Cat, [224 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(225, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 215, [225 | __Ss], [__T | __Stack]);
yeccpars2(225, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(226, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 215, [226 | __Ss], [__T | __Stack]);
yeccpars2(226, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(227, 'RSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 228, [227 | __Ss], [__T | __Stack]);
yeccpars2(227, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(228, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'PropertyParm'{value = [__2|__3], extraInfo = {sublist,true}},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(alternativeValue, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(229, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 226, [229 | __Ss], [__T | __Stack]);
yeccpars2(229, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(230, __Cat, [229 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(230, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(valueList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(231, 'RSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 232, [231 | __Ss], [__T | __Stack]);
yeccpars2(231, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(232, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'PropertyParm'{value = [__2,__4], extraInfo = {range,true}},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(alternativeValue, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(233, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 226, [233 | __Ss], [__T | __Stack]);
yeccpars2(233, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(234, __Cat, [233 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(234, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 235, [234 | __Ss], [__T | __Stack]);
yeccpars2(234, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(235, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'PropertyParm'{value = [__2|__3], extraInfo = {sublist,false}},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(alternativeValue, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(236, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [236 | __Ss], [__T | __Stack]);
yeccpars2(236, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(237, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_streamID(__1),
 yeccpars2(yeccgoto(streamID, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(238, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {stream,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(sigParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(239, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 240, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 242, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 241, [239 | __Ss], [__T | __Stack]);
yeccpars2(239, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(240, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  brief,
 yeccpars2(yeccgoto(signalType, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(241, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  onOff,
 yeccpars2(yeccgoto(signalType, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(242, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  timeOut,
 yeccpars2(yeccgoto(signalType, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(243, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {signal_type,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(sigParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(244, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 245, [244 | __Ss], [__T | __Stack]);
yeccpars2(244, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(245, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 248, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 247, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 246, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 249, [245 | __Ss], [__T | __Stack]);
yeccpars2(245, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(246, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  onInterruptByEvent,
 yeccpars2(yeccgoto(notificationReason, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(247, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  onInterruptByNewSignalDescr,
 yeccpars2(yeccgoto(notificationReason, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(248, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  otherReason,
 yeccpars2(yeccgoto(notificationReason, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(249, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  onTimeOut,
 yeccpars2(yeccgoto(notificationReason, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(250, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 251, [250 | __Ss], [__T | __Stack]);
yeccpars2(250, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(252, __Cat, [250 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(251, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 248, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 247, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 246, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 249, [251 | __Ss], [__T | __Stack]);
yeccpars2(251, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(252, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 253, [252 | __Ss], [__T | __Stack]);
yeccpars2(252, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(253, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {notify_completion,[__4|__5]},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(sigParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(254, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 251, [254 | __Ss], [__T | __Stack]);
yeccpars2(254, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(255, __Cat, [254 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(255, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(notificationReasons, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(256, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [256 | __Ss], [__T | __Stack]);
yeccpars2(256, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(257, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {duration,ensure_uint16(__3)},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(sigParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(258, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [258 | __Ss], [__T | __Stack]);
yeccpars2(258, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(259, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_uint16(__1),
 yeccpars2(yeccgoto(signalListId, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(260, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 261, [260 | __Ss], [__T | __Stack]);
yeccpars2(260, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(261, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [261 | __Ss], [__T | __Stack]);
yeccpars2(261, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(262, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 264, [262 | __Ss], [__T | __Stack]);
yeccpars2(262, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(265, __Cat, [262 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(263, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(signalListParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(264, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [264 | __Ss], [__T | __Stack]);
yeccpars2(264, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(265, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 266, [265 | __Ss], [__T | __Stack]);
yeccpars2(265, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(266, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'SeqSigList'{id = ensure_uint16(__3), signalList = [__5|__6]},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto(signalList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(267, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 264, [267 | __Ss], [__T | __Stack]);
yeccpars2(267, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(268, __Cat, [267 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(268, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(signalListParms, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(269, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [269 | __Ss], [__T | __Stack]);
yeccpars2(269, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(270, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 272, [270 | __Ss], [__T | __Stack]);
yeccpars2(270, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(271, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_muxType(__1),
 yeccpars2(yeccgoto(muxType, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(272, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [272 | __Ss], [__T | __Stack]);
yeccpars2(272, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(273, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'MuxDescriptor'{muxType = __3, termList = __4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(muxDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(274, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 275, [274 | __Ss], [__T | __Stack]);
yeccpars2(274, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(276, __Cat, [274 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(275, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [275 | __Ss], [__T | __Stack]);
yeccpars2(275, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(276, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 277, [276 | __Ss], [__T | __Stack]);
yeccpars2(276, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(277, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(terminationIDList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(278, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 275, [278 | __Ss], [__T | __Stack]);
yeccpars2(278, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(279, __Cat, [278 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(279, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(terminationIDListRepeat, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(280, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [280 | __Ss], [__T | __Stack]);
yeccpars2(280, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(281, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [281 | __Ss], [__T | __Stack]);
yeccpars2(281, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(282, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 284, [282 | __Ss], [__T | __Stack]);
yeccpars2(282, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(285, __Cat, [282 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(283, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_modemType(__1),
 yeccpars2(yeccgoto(modemType, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(284, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [284 | __Ss], [__T | __Stack]);
yeccpars2(284, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(285, 'RSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 286, [285 | __Ss], [__T | __Stack]);
yeccpars2(285, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(286, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 287, [286 | __Ss], [__T | __Stack]);
yeccpars2(286, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(288, __Cat, [286 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(287, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [287 | __Ss], [__T | __Stack]);
yeccpars2(287, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(288, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'ModemDescriptor'{mtl = [__3|__4], mpl = __6},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(modemDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(289, 'GREATER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 211, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'LESSER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 212, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'NEQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 213, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 210, [289 | __Ss], [__T | __Stack]);
yeccpars2(289, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(290, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 291, [290 | __Ss], [__T | __Stack]);
yeccpars2(290, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(292, __Cat, [290 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(291, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [291 | __Ss], [__T | __Stack]);
yeccpars2(291, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(292, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 293, [292 | __Ss], [__T | __Stack]);
yeccpars2(292, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(293, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(optPropertyParms, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(294, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 291, [294 | __Ss], [__T | __Stack]);
yeccpars2(294, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(295, __Cat, [294 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(295, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(propertyParms, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(296, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  setelement(#'PropertyParm'.name,__2,__1),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(propertyParm, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(297, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 284, [297 | __Ss], [__T | __Stack]);
yeccpars2(297, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(298, __Cat, [297 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(298, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(modemTypeList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(299, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 287, [299 | __Ss], [__T | __Stack]);
yeccpars2(299, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(300, __Cat, [299 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(300, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'ModemDescriptor'{mtl = [__3], mpl = __4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(modemDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(301, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 306, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 302, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 305, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, 'RemoteDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 304, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, 'LocalDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 303, [301 | __Ss], [__T | __Stack]);
yeccpars2(301, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(302, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 347, [302 | __Ss], [__T | __Stack]);
yeccpars2(302, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(303, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {local,#'LocalRemoteDescriptor'{propGrps = ensure_prop_groups(__1)}},
 yeccpars2(yeccgoto(streamParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(304, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {remote,#'LocalRemoteDescriptor'{propGrps = ensure_prop_groups(__1)}},
 yeccpars2(yeccgoto(streamParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(305, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 338, [305 | __Ss], [__T | __Stack]);
yeccpars2(305, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(306, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 317, [306 | __Ss], [__T | __Stack]);
yeccpars2(306, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(307, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {control,__1},
 yeccpars2(yeccgoto(streamParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(308, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 312, [308 | __Ss], [__T | __Stack]);
yeccpars2(308, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(313, __Cat, [308 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(309, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {streamDescriptor,__1},
 yeccpars2(yeccgoto(mediaParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(310, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {streamParm,__1},
 yeccpars2(yeccgoto(mediaParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(311, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {termState,__1},
 yeccpars2(yeccgoto(mediaParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(312, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 306, [312 | __Ss], [__T | __Stack]);
yeccpars2(312, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 302, [312 | __Ss], [__T | __Stack]);
yeccpars2(312, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 305, [312 | __Ss], [__T | __Stack]);
yeccpars2(312, 'RemoteDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 304, [312 | __Ss], [__T | __Stack]);
yeccpars2(312, 'LocalDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 303, [312 | __Ss], [__T | __Stack]);
yeccpars2(312, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(313, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 314, [313 | __Ss], [__T | __Stack]);
yeccpars2(313, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(314, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_mediaDescriptor([__3|__4]),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(mediaDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(315, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 312, [315 | __Ss], [__T | __Stack]);
yeccpars2(315, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(316, __Cat, [315 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(316, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(mediaParmList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(317, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 319, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 318, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [317 | __Ss], [__T | __Stack]);
yeccpars2(317, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(318, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 334, [318 | __Ss], [__T | __Stack]);
yeccpars2(318, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(319, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 329, [319 | __Ss], [__T | __Stack]);
yeccpars2(319, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(320, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {eventBufferControl,__1},
 yeccpars2(yeccgoto(terminationStateParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(321, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {propertyParm,__1},
 yeccpars2(yeccgoto(terminationStateParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(322, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {serviceState,__1},
 yeccpars2(yeccgoto(terminationStateParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(323, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 324, [323 | __Ss], [__T | __Stack]);
yeccpars2(323, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(325, __Cat, [323 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(324, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 319, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 318, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [324 | __Ss], [__T | __Stack]);
yeccpars2(324, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(325, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 326, [325 | __Ss], [__T | __Stack]);
yeccpars2(325, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(326, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_terminationStateDescriptor([__3|__4]),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(terminationStateDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(327, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 324, [327 | __Ss], [__T | __Stack]);
yeccpars2(327, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(328, __Cat, [327 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(328, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(terminationStateParms, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(329, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 330, [329 | __Ss], [__T | __Stack]);
yeccpars2(329, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 331, [329 | __Ss], [__T | __Stack]);
yeccpars2(329, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 332, [329 | __Ss], [__T | __Stack]);
yeccpars2(329, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(330, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  inSvc,
 yeccpars2(yeccgoto(serviceState, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(331, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  outOfSvc,
 yeccpars2(yeccgoto(serviceState, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(332, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  test,
 yeccpars2(yeccgoto(serviceState, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(333, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceStates, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(334, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 335, [334 | __Ss], [__T | __Stack]);
yeccpars2(334, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 336, [334 | __Ss], [__T | __Stack]);
yeccpars2(334, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(335, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  lockStep,
 yeccpars2(yeccgoto(eventBufferControlState, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(336, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  off,
 yeccpars2(yeccgoto(eventBufferControlState, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(337, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(eventBufferControl, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(338, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [338 | __Ss], [__T | __Stack]);
yeccpars2(338, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(339, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 340, [339 | __Ss], [__T | __Stack]);
yeccpars2(339, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(340, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 302, [340 | __Ss], [__T | __Stack]);
yeccpars2(340, 'RemoteDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 304, [340 | __Ss], [__T | __Stack]);
yeccpars2(340, 'LocalDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 303, [340 | __Ss], [__T | __Stack]);
yeccpars2(340, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(341, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 342, [341 | __Ss], [__T | __Stack]);
yeccpars2(341, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(343, __Cat, [341 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(342, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 302, [342 | __Ss], [__T | __Stack]);
yeccpars2(342, 'RemoteDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 304, [342 | __Ss], [__T | __Stack]);
yeccpars2(342, 'LocalDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 303, [342 | __Ss], [__T | __Stack]);
yeccpars2(342, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(343, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 344, [343 | __Ss], [__T | __Stack]);
yeccpars2(343, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(344, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'StreamDescriptor'{streamID = __3, streamParms = merge_streamParms([__5|__6])},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto(streamDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(345, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 342, [345 | __Ss], [__T | __Stack]);
yeccpars2(345, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(346, __Cat, [345 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(346, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(streamParmList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(347, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 350, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 349, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 348, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [347 | __Ss], [__T | __Stack]);
yeccpars2(347, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(348, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 364, [348 | __Ss], [__T | __Stack]);
yeccpars2(348, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(349, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 362, [349 | __Ss], [__T | __Stack]);
yeccpars2(349, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(350, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 358, [350 | __Ss], [__T | __Stack]);
yeccpars2(350, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(351, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 353, [351 | __Ss], [__T | __Stack]);
yeccpars2(351, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(354, __Cat, [351 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(352, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {prop,__1},
 yeccpars2(yeccgoto(localParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(353, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 350, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 349, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 348, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [353 | __Ss], [__T | __Stack]);
yeccpars2(353, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(354, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 355, [354 | __Ss], [__T | __Stack]);
yeccpars2(354, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(355, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(localControlDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(356, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 353, [356 | __Ss], [__T | __Stack]);
yeccpars2(356, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(357, __Cat, [356 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(357, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(localParmList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(358, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 359, [358 | __Ss], [__T | __Stack]);
yeccpars2(358, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 360, [358 | __Ss], [__T | __Stack]);
yeccpars2(358, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(359, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  false,
 yeccpars2(yeccgoto(onOrOff, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(360, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  true,
 yeccpars2(yeccgoto(onOrOff, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(361, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {value,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(localParm, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(362, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 359, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 360, [362 | __Ss], [__T | __Stack]);
yeccpars2(362, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(363, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {group,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(localParm, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(364, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 366, [364 | __Ss], [__T | __Stack]);
yeccpars2(364, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 365, [364 | __Ss], [__T | __Stack]);
yeccpars2(364, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 369, [364 | __Ss], [__T | __Stack]);
yeccpars2(364, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 367, [364 | __Ss], [__T | __Stack]);
yeccpars2(364, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 368, [364 | __Ss], [__T | __Stack]);
yeccpars2(364, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(365, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  inactive,
 yeccpars2(yeccgoto(streamModes, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(366, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  loopBack,
 yeccpars2(yeccgoto(streamModes, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(367, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  recvOnly,
 yeccpars2(yeccgoto(streamModes, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(368, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  sendOnly,
 yeccpars2(yeccgoto(streamModes, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(369, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  sendRecv,
 yeccpars2(yeccgoto(streamModes, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(370, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {mode,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(localParm, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(371, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [371 | __Ss], [__T | __Stack]);
yeccpars2(371, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(372, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 374, [372 | __Ss], [__T | __Stack]);
yeccpars2(372, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(373, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_requestID(__1),
 yeccpars2(yeccgoto(requestID, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(374, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [374 | __Ss], [__T | __Stack]);
yeccpars2(374, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(375, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 382, [375 | __Ss], [__T | __Stack]);
yeccpars2(375, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  #'RequestedEvent'{evParList = []},
 yeccpars2(383, __Cat, [375 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(376, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 377, [376 | __Ss], [__T | __Stack]);
yeccpars2(376, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(378, __Cat, [376 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(377, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [377 | __Ss], [__T | __Stack]);
yeccpars2(377, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(378, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 379, [378 | __Ss], [__T | __Stack]);
yeccpars2(378, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(379, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'EventsDescriptor'{requestID = __3, eventList = [__5|__6]},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto(eventsDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(380, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 377, [380 | __Ss], [__T | __Stack]);
yeccpars2(380, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(381, __Cat, [380 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(381, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(requestedEvents, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(382, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 386, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 385, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 384, [382 | __Ss], [__T | __Stack]);
yeccpars2(382, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(383, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  setelement(#'RequestedEvent'.pkgdName,__2,__1),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(requestedEvent, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(384, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_eventDM(__1),
 yeccpars2(yeccgoto(eventDM, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(385, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 400, [385 | __Ss], [__T | __Stack]);
yeccpars2(385, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(386, 'COMMA', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  keepActive,
 yeccpars2(yeccgoto(eventParameter, hd(__Ss)), 'COMMA', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(386, 'RBRKT', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  keepActive,
 yeccpars2(yeccgoto(eventParameter, hd(__Ss)), 'RBRKT', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(386, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(387, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(eventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(388, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(eventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(389, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(eventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(390, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 395, [390 | __Ss], [__T | __Stack]);
yeccpars2(390, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(396, __Cat, [390 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(391, 'GREATER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 211, [391 | __Ss], [__T | __Stack]);
yeccpars2(391, 'LESSER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 212, [391 | __Ss], [__T | __Stack]);
yeccpars2(391, 'NEQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 213, [391 | __Ss], [__T | __Stack]);
yeccpars2(391, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 210, [391 | __Ss], [__T | __Stack]);
yeccpars2(391, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(392, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(eventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(393, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_NAME(__1),
 yeccpars2(yeccgoto(eventParameterName, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(394, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  select_stream_or_other(__1,__2),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(eventStreamOrOther, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(395, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 386, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 385, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 384, [395 | __Ss], [__T | __Stack]);
yeccpars2(395, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(396, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 397, [396 | __Ss], [__T | __Stack]);
yeccpars2(396, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(397, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_eventParameters([__2|__3]),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(requestedEventBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(398, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 395, [398 | __Ss], [__T | __Stack]);
yeccpars2(398, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(399, __Cat, [398 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(399, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(eventParameters, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(400, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 169, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 401, [400 | __Ss], [__T | __Stack]);
yeccpars2(400, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(401, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 409, [401 | __Ss], [__T | __Stack]);
yeccpars2(401, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'SecondEventsDescriptor'{requestID = asn1_NOVALUE, eventList = []},
 yeccpars2(yeccgoto(embedFirst, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(402, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 408, [402 | __Ss], [__T | __Stack]);
yeccpars2(402, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(403, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 405, [403 | __Ss], [__T | __Stack]);
yeccpars2(403, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 404, [403 | __Ss], [__T | __Stack]);
yeccpars2(403, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(404, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 401, [404 | __Ss], [__T | __Stack]);
yeccpars2(404, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(405, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {embed,__3,asn1_NOVALUE},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(embedWithSig, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(406, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 407, [406 | __Ss], [__T | __Stack]);
yeccpars2(406, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(407, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {embed,__3,__5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(embedWithSig, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(408, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {embed,asn1_NOVALUE,__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(embedNoSig, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(409, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [409 | __Ss], [__T | __Stack]);
yeccpars2(409, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(410, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 411, [410 | __Ss], [__T | __Stack]);
yeccpars2(410, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(411, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [411 | __Ss], [__T | __Stack]);
yeccpars2(411, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(412, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 419, [412 | __Ss], [__T | __Stack]);
yeccpars2(412, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  #'SecondRequestedEvent'{evParList = []},
 yeccpars2(420, __Cat, [412 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(413, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 414, [413 | __Ss], [__T | __Stack]);
yeccpars2(413, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(415, __Cat, [413 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(414, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [414 | __Ss], [__T | __Stack]);
yeccpars2(414, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(415, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 416, [415 | __Ss], [__T | __Stack]);
yeccpars2(415, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(416, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'SecondEventsDescriptor'{requestID = __3, eventList = [__5|__6]},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto(embedFirst, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(417, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 414, [417 | __Ss], [__T | __Stack]);
yeccpars2(417, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(418, __Cat, [417 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(418, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(secondRequestedEvents, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(419, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 422, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 421, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 384, [419 | __Ss], [__T | __Stack]);
yeccpars2(419, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(420, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  setelement(#'SecondRequestedEvent'.pkgdName,__2,__1),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(secondRequestedEvent, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(421, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 432, [421 | __Ss], [__T | __Stack]);
yeccpars2(421, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(422, 'COMMA', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  keepActive,
 yeccpars2(yeccgoto(secondEventParameter, hd(__Ss)), 'COMMA', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(422, 'RBRKT', __Ss, [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  keepActive,
 yeccpars2(yeccgoto(secondEventParameter, hd(__Ss)), 'RBRKT', __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(422, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(423, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(secondEventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(424, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(secondEventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(425, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(secondEventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(426, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 427, [426 | __Ss], [__T | __Stack]);
yeccpars2(426, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(428, __Cat, [426 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(427, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 422, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 421, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 384, [427 | __Ss], [__T | __Stack]);
yeccpars2(427, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(428, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 429, [428 | __Ss], [__T | __Stack]);
yeccpars2(428, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(429, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_secondEventParameters([__2|__3]),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(secondRequestedEventBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(430, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 427, [430 | __Ss], [__T | __Stack]);
yeccpars2(430, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(431, __Cat, [430 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(431, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(secondEventParameters, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(432, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 169, [432 | __Ss], [__T | __Stack]);
yeccpars2(432, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(433, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 434, [433 | __Ss], [__T | __Stack]);
yeccpars2(433, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(434, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {second_embed,__3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(embedSig, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(435, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 436, [435 | __Ss], [__T | __Stack]);
yeccpars2(435, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [435 | __Ss], [__T | __Stack]);
yeccpars2(435, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(439, __Cat, [435 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(436, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_timeStamp(__1),
 yeccpars2(yeccgoto(timeStamp, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(437, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 456, [437 | __Ss], [__T | __Stack]);
yeccpars2(437, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(457, __Cat, [437 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(438, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_eventSpec(__1),
 yeccpars2(yeccgoto(eventSpec, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(439, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [439 | __Ss], [__T | __Stack]);
yeccpars2(439, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(440, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [440 | __Ss], [__T | __Stack]);
yeccpars2(440, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(441, __Cat, [440 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(441, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 442, [441 | __Ss], [__T | __Stack]);
yeccpars2(441, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(442, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [442 | __Ss], [__T | __Stack]);
yeccpars2(442, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(443, __Cat, [442 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(443, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [443 | __Ss], [__T | __Stack]);
yeccpars2(443, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(444, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 445, [444 | __Ss], [__T | __Stack]);
yeccpars2(444, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(446, __Cat, [444 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(445, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [445 | __Ss], [__T | __Stack]);
yeccpars2(445, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(446, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_observed_event(__6,__5,__1),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(observedEvent, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(447, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(observedEventParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(448, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 449, [448 | __Ss], [__T | __Stack]);
yeccpars2(448, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(450, __Cat, [448 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(449, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [449 | __Ss], [__T | __Stack]);
yeccpars2(449, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(450, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 451, [450 | __Ss], [__T | __Stack]);
yeccpars2(450, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(451, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(observedEventBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(452, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 449, [452 | __Ss], [__T | __Stack]);
yeccpars2(452, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(453, __Cat, [452 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(453, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(observedEventParameters, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(454, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 445, [454 | __Ss], [__T | __Stack]);
yeccpars2(454, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(455, __Cat, [454 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(455, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_observed_event(__3,__2,asn1_NOVALUE),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(observedEvent, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(456, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 436, [456 | __Ss], [__T | __Stack]);
yeccpars2(456, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [456 | __Ss], [__T | __Stack]);
yeccpars2(456, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(439, __Cat, [456 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(457, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 458, [457 | __Ss], [__T | __Stack]);
yeccpars2(457, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(458, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(eventBufferDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(459, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 456, [459 | __Ss], [__T | __Stack]);
yeccpars2(459, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(460, __Cat, [459 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(460, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(eventSpecList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(461, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 470, [461 | __Ss], [__T | __Stack]);
yeccpars2(461, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 472, [461 | __Ss], [__T | __Stack]);
yeccpars2(461, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 462, [461 | __Ss], [__T | __Stack]);
yeccpars2(461, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 471, [461 | __Ss], [__T | __Stack]);
yeccpars2(461, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 465, [461 | __Ss], [__T | __Stack]);
yeccpars2(461, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 464, [461 | __Ss], [__T | __Stack]);
yeccpars2(461, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 466, [461 | __Ss], [__T | __Stack]);
yeccpars2(461, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 469, [461 | __Ss], [__T | __Stack]);
yeccpars2(461, 'DigitMapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 463, [461 | __Ss], [__T | __Stack]);
yeccpars2(461, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 467, [461 | __Ss], [__T | __Stack]);
yeccpars2(461, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 468, [461 | __Ss], [__T | __Stack]);
yeccpars2(461, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(473, __Cat, [461 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(462, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_IADMD(__1),
 yeccpars2(yeccgoto(indAuddigitMapDescriptor, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(463, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  digitMapToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(464, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 544, [464 | __Ss], [__T | __Stack]);
yeccpars2(464, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  eventBufferToken,
 yeccpars2(yeccgoto(auditItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(465, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 538, [465 | __Ss], [__T | __Stack]);
yeccpars2(465, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  eventsToken,
 yeccpars2(yeccgoto(auditItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(466, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 515, [466 | __Ss], [__T | __Stack]);
yeccpars2(466, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mediaToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(467, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  modemToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(468, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  muxToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(469, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  observedEventsToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(470, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 511, [470 | __Ss], [__T | __Stack]);
yeccpars2(470, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  packagesToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(471, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 498, [471 | __Ss], [__T | __Stack]);
yeccpars2(471, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  signalsToken,
 yeccpars2(yeccgoto(auditItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(472, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 495, [472 | __Ss], [__T | __Stack]);
yeccpars2(472, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  statsToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(473, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 561, [473 | __Ss], [__T | __Stack]);
yeccpars2(473, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(474, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 557, [474 | __Ss], [__T | __Stack]);
yeccpars2(474, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(558, __Cat, [474 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(475, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(auditItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(476, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 485, [476 | __Ss], [__T | __Stack]);
yeccpars2(476, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(486, __Cat, [476 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(477, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {indAudDigitMapDescriptor,__1},
 yeccpars2(yeccgoto(indAudauditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(478, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {indAudEventBufferDescriptor,__1},
 yeccpars2(yeccgoto(indAudauditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(479, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {indAudEventsDescriptor,__1},
 yeccpars2(yeccgoto(indAudauditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(480, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {indAudMediaDescriptor,__1},
 yeccpars2(yeccgoto(indAudauditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(481, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {indAudPackagesDescriptor,__1},
 yeccpars2(yeccgoto(indAudauditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(482, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {indAudSignalsDescriptor,__1},
 yeccpars2(yeccgoto(indAudauditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(483, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {indAudStatisticsDescriptor,__1},
 yeccpars2(yeccgoto(indAudauditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(484, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {terminationAudit,__1},
 yeccpars2(yeccgoto(auditItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(485, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 490, [485 | __Ss], [__T | __Stack]);
yeccpars2(485, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 492, [485 | __Ss], [__T | __Stack]);
yeccpars2(485, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 462, [485 | __Ss], [__T | __Stack]);
yeccpars2(485, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 491, [485 | __Ss], [__T | __Stack]);
yeccpars2(485, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 488, [485 | __Ss], [__T | __Stack]);
yeccpars2(485, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 487, [485 | __Ss], [__T | __Stack]);
yeccpars2(485, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 489, [485 | __Ss], [__T | __Stack]);
yeccpars2(485, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(486, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(indAudterminationAudit, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(487, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 544, [487 | __Ss], [__T | __Stack]);
yeccpars2(487, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(488, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 538, [488 | __Ss], [__T | __Stack]);
yeccpars2(488, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(489, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 515, [489 | __Ss], [__T | __Stack]);
yeccpars2(489, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(490, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 511, [490 | __Ss], [__T | __Stack]);
yeccpars2(490, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(491, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 498, [491 | __Ss], [__T | __Stack]);
yeccpars2(491, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(492, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 495, [492 | __Ss], [__T | __Stack]);
yeccpars2(492, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(493, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 485, [493 | __Ss], [__T | __Stack]);
yeccpars2(493, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(494, __Cat, [493 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(494, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(indAudterminationAuditList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(495, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [495 | __Ss], [__T | __Stack]);
yeccpars2(495, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(496, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 497, [496 | __Ss], [__T | __Stack]);
yeccpars2(496, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(497, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'IndAudStatisticsDescriptor'{statName = __3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(indAudstatisticsDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(498, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 501, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 500, [498 | __Ss], [__T | __Stack]);
yeccpars2(498, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(499, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(indAudsignalsDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(500, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(optIndAudsignalParm, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(501, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 506, [501 | __Ss], [__T | __Stack]);
yeccpars2(501, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(502, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {seqSigList,__1},
 yeccpars2(yeccgoto(indAudsignalParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(503, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 505, [503 | __Ss], [__T | __Stack]);
yeccpars2(503, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(504, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {signal,ensure_indAudSignal(__1)},
 yeccpars2(yeccgoto(indAudsignalParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(505, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(optIndAudsignalParm, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(506, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [506 | __Ss], [__T | __Stack]);
yeccpars2(506, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(507, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 508, [507 | __Ss], [__T | __Stack]);
yeccpars2(507, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(508, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [508 | __Ss], [__T | __Stack]);
yeccpars2(508, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(509, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 510, [509 | __Ss], [__T | __Stack]);
yeccpars2(509, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(510, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'IndAudSeqSigList'{id = ensure_uint16(__3), signalList = ensure_indAudSignalListParm(__5)},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(indAudsignalList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(511, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [511 | __Ss], [__T | __Stack]);
yeccpars2(511, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(512, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 514, [512 | __Ss], [__T | __Stack]);
yeccpars2(512, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(513, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_packagesItem(__1),
 yeccpars2(yeccgoto(packagesItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(514, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_indAudPackagesDescriptor(__3),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(indAudpackagesDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(515, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 518, [515 | __Ss], [__T | __Stack]);
yeccpars2(515, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 516, [515 | __Ss], [__T | __Stack]);
yeccpars2(515, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 517, [515 | __Ss], [__T | __Stack]);
yeccpars2(515, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(516, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 534, [516 | __Ss], [__T | __Stack]);
yeccpars2(516, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(517, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 529, [517 | __Ss], [__T | __Stack]);
yeccpars2(517, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(518, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 525, [518 | __Ss], [__T | __Stack]);
yeccpars2(518, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(519, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'IndAudStreamParms'{localControlDescriptor = __1},
 yeccpars2(yeccgoto(indAudstreamParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(520, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 524, [520 | __Ss], [__T | __Stack]);
yeccpars2(520, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(521, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {streamDescr,__1},
 yeccpars2(yeccgoto(indAudmediaParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(522, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {streamParm,__1},
 yeccpars2(yeccgoto(indAudmediaParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(523, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {termStateDescr,__1},
 yeccpars2(yeccgoto(indAudmediaParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(524, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_indAudMediaDescriptor(__3),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(indAudmediaDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(525, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [525 | __Ss], [__T | __Stack]);
yeccpars2(525, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(526, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 528, [526 | __Ss], [__T | __Stack]);
yeccpars2(526, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(527, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_indAudTerminationStateParm(__1),
 yeccpars2(yeccgoto(indAudterminationStateParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(528, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_indAudTerminationStateDescriptor(__3),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(indAudterminationStateDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(529, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [529 | __Ss], [__T | __Stack]);
yeccpars2(529, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(530, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 531, [530 | __Ss], [__T | __Stack]);
yeccpars2(530, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(531, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 516, [531 | __Ss], [__T | __Stack]);
yeccpars2(531, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(532, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 533, [532 | __Ss], [__T | __Stack]);
yeccpars2(532, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(533, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'IndAudStreamDescriptor'{streamID = __3, streamParms = __5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(indAudstreamDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(534, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [534 | __Ss], [__T | __Stack]);
yeccpars2(534, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(535, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 537, [535 | __Ss], [__T | __Stack]);
yeccpars2(535, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(536, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_indAudLocalParm(__1),
 yeccpars2(yeccgoto(indAudlocalParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(537, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_indAudLocalControlDescriptor(__3),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(indAudlocalControlDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(538, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [538 | __Ss], [__T | __Stack]);
yeccpars2(538, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(539, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 540, [539 | __Ss], [__T | __Stack]);
yeccpars2(539, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(540, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [540 | __Ss], [__T | __Stack]);
yeccpars2(540, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(541, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 543, [541 | __Ss], [__T | __Stack]);
yeccpars2(541, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(542, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(indAudrequestedEvent, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(543, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'IndAudEventsDescriptor'{requestID = __3, pkgdName = __5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(indAudeventsDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(544, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [544 | __Ss], [__T | __Stack]);
yeccpars2(544, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(545, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 556, [545 | __Ss], [__T | __Stack]);
yeccpars2(545, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(546, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 547, [546 | __Ss], [__T | __Stack]);
yeccpars2(546, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(548, __Cat, [546 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(547, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 549, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [547 | __Ss], [__T | __Stack]);
yeccpars2(547, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(548, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_indAudEventBufferDescriptor(__1,__2),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(indAudeventSpec, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(549, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 554, [549 | __Ss], [__T | __Stack]);
yeccpars2(549, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(550, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {eventParameterName,__1},
 yeccpars2(yeccgoto(indAudeventSpecParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(551, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {streamID,__1},
 yeccpars2(yeccgoto(indAudeventSpecParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(552, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 553, [552 | __Ss], [__T | __Stack]);
yeccpars2(552, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(553, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(optIndAudeventSpecParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(554, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [554 | __Ss], [__T | __Stack]);
yeccpars2(554, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(555, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(eventStream, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(556, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(indAudeventBufferDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(557, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 470, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 472, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 462, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 471, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 465, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 464, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 466, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 469, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'DigitMapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 463, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 467, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 468, [557 | __Ss], [__T | __Stack]);
yeccpars2(557, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(558, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(auditDescriptorBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(559, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 557, [559 | __Ss], [__T | __Stack]);
yeccpars2(559, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(560, __Cat, [559 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(560, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(auditItemList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(561, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_auditDescriptor(__3),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(auditDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(562, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 142, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 140, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 134, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 135, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 143, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 138, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 139, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 133, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 136, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [562 | __Ss], [__T | __Stack]);
yeccpars2(562, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(563, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 564, [563 | __Ss], [__T | __Stack]);
yeccpars2(563, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(564, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_action_requests(__3,[__5|__6]),
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto(actionRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(565, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 562, [565 | __Ss], [__T | __Stack]);
yeccpars2(565, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(566, __Cat, [565 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(566, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(actionRequestItems, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(567, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [567 | __Ss], [__T | __Stack]);
yeccpars2(567, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(568, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 576, [568 | __Ss], [__T | __Stack]);
yeccpars2(568, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(569, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(terminationA, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(570, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 571, [570 | __Ss], [__T | __Stack]);
yeccpars2(570, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(572, __Cat, [570 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(571, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [571 | __Ss], [__T | __Stack]);
yeccpars2(571, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(572, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 573, [572 | __Ss], [__T | __Stack]);
yeccpars2(572, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(573, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(topologyDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(574, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 571, [574 | __Ss], [__T | __Stack]);
yeccpars2(574, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(575, __Cat, [574 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(575, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(topologyTripleList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(576, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [576 | __Ss], [__T | __Stack]);
yeccpars2(576, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(577, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 579, [577 | __Ss], [__T | __Stack]);
yeccpars2(577, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(578, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __1,
 yeccpars2(yeccgoto(terminationB, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(579, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 582, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 581, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 580, [579 | __Ss], [__T | __Stack]);
yeccpars2(579, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(580, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  bothway,
 yeccpars2(yeccgoto(topologyDirection, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(581, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  isolate,
 yeccpars2(yeccgoto(topologyDirection, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(582, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  oneway,
 yeccpars2(yeccgoto(topologyDirection, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(583, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'TopologyRequest'{terminationFrom = __1, terminationTo = __3, topologyDirection = __5},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(topologyTriple, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(584, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [584 | __Ss], [__T | __Stack]);
yeccpars2(584, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(585, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 586, [585 | __Ss], [__T | __Stack]);
yeccpars2(585, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(587, __Cat, [585 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(586, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 162, [586 | __Ss], [__T | __Stack]);
yeccpars2(586, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(587, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_commandRequest({subtractReq,__1},#'SubtractRequest'{terminationID = [__3], auditDescriptor = __4}),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(subtractRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(588, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 589, [588 | __Ss], [__T | __Stack]);
yeccpars2(588, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(589, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(optAuditDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(590, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [590 | __Ss], [__T | __Stack]);
yeccpars2(590, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(591, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 592, [591 | __Ss], [__T | __Stack]);
yeccpars2(591, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(592, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 593, [592 | __Ss], [__T | __Stack]);
yeccpars2(592, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(593, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 596, [593 | __Ss], [__T | __Stack]);
yeccpars2(593, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(594, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 595, [594 | __Ss], [__T | __Stack]);
yeccpars2(594, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(595, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_commandRequest({serviceChangeReq,__1},#'ServiceChangeRequest'{terminationID = [__3], serviceChangeParms = __5}),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(serviceChangeRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(596, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 603, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 602, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 601, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 600, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 599, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 598, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 597, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 436, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 470, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 472, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 462, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 471, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 465, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 464, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 466, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 469, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'DigitMapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 463, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 467, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 468, [596 | __Ss], [__T | __Stack]);
yeccpars2(596, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(597, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 637, [597 | __Ss], [__T | __Stack]);
yeccpars2(597, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(598, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 635, [598 | __Ss], [__T | __Stack]);
yeccpars2(598, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(599, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 633, [599 | __Ss], [__T | __Stack]);
yeccpars2(599, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(600, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 631, [600 | __Ss], [__T | __Stack]);
yeccpars2(600, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(601, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 629, [601 | __Ss], [__T | __Stack]);
yeccpars2(601, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(602, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 625, [602 | __Ss], [__T | __Stack]);
yeccpars2(602, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(603, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 623, [603 | __Ss], [__T | __Stack]);
yeccpars2(603, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(604, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {audit_item,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(605, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {extension,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(606, 'GREATER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 211, [606 | __Ss], [__T | __Stack]);
yeccpars2(606, 'LESSER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 212, [606 | __Ss], [__T | __Stack]);
yeccpars2(606, 'NEQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 213, [606 | __Ss], [__T | __Stack]);
yeccpars2(606, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 210, [606 | __Ss], [__T | __Stack]);
yeccpars2(606, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(607, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_extensionParameter(__1),
 yeccpars2(yeccgoto(extensionParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(608, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {address,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(609, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {delay,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(610, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {method,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(611, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {mgc_id,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(612, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 617, [612 | __Ss], [__T | __Stack]);
yeccpars2(612, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(618, __Cat, [612 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(613, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {profile,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(614, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {reason,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(615, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {version,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(616, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {time_stamp,__1},
 yeccpars2(yeccgoto(serviceChangeParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(617, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 603, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 602, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 601, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 600, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 599, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 598, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 597, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 436, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 470, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 472, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 462, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 471, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 465, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 464, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 466, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 469, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'DigitMapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 463, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 467, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 468, [617 | __Ss], [__T | __Stack]);
yeccpars2(617, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(618, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 619, [618 | __Ss], [__T | __Stack]);
yeccpars2(618, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(619, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_ServiceChangeParm([__3|__4]),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(serviceChangeDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(620, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 617, [620 | __Ss], [__T | __Stack]);
yeccpars2(620, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(621, __Cat, [620 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(621, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeParms, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(622, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  setelement(#'PropertyParm'.name,__2,__1),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(extension, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(623, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [623 | __Ss], [__T | __Stack]);
yeccpars2(623, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(624, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_version(__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeVersion, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(625, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'LSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'LESSER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 93, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [625 | __Ss], [__T | __Stack]);
yeccpars2(625, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(98, __Cat, [625 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(626, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeAddress, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(627, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {portNumber,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeAddress, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(628, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_uint16(__1),
 yeccpars2(yeccgoto(portNumber, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(629, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 215, [629 | __Ss], [__T | __Stack]);
yeccpars2(629, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(630, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeReason, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(631, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [631 | __Ss], [__T | __Stack]);
yeccpars2(631, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(632, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_profile(__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeProfile, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(633, 'LSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 94, [633 | __Ss], [__T | __Stack]);
yeccpars2(633, 'LESSER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 93, [633 | __Ss], [__T | __Stack]);
yeccpars2(633, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [633 | __Ss], [__T | __Stack]);
yeccpars2(633, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(98, __Cat, [633 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(634, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __3,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeMgcId, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(635, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [635 | __Ss], [__T | __Stack]);
yeccpars2(635, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(636, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_serviceChangeMethod(__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeMethod, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(637, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [637 | __Ss], [__T | __Stack]);
yeccpars2(637, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(638, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_uint32(__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeDelay, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(639, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [639 | __Ss], [__T | __Stack]);
yeccpars2(639, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(640, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_uint16(__3),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(priority, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(641, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [641 | __Ss], [__T | __Stack]);
yeccpars2(641, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(642, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 643, [642 | __Ss], [__T | __Stack]);
yeccpars2(642, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(643, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 644, [643 | __Ss], [__T | __Stack]);
yeccpars2(643, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [643 | __Ss], [__T | __Stack]);
yeccpars2(643, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(644, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 649, [644 | __Ss], [__T | __Stack]);
yeccpars2(644, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(645, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'NotifyRequest'{errorDescriptor = __1},
 yeccpars2(yeccgoto(notifyRequestBody, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(646, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 648, [646 | __Ss], [__T | __Stack]);
yeccpars2(646, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(647, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'NotifyRequest'{observedEventsDescriptor = __1},
 yeccpars2(yeccgoto(notifyRequestBody, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(648, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_commandRequest({notifyReq,__1},setelement(#'NotifyRequest'.terminationID,__5,[__3])),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(notifyRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(649, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [649 | __Ss], [__T | __Stack]);
yeccpars2(649, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(650, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 651, [650 | __Ss], [__T | __Stack]);
yeccpars2(650, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(651, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 436, [651 | __Ss], [__T | __Stack]);
yeccpars2(651, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [651 | __Ss], [__T | __Stack]);
yeccpars2(651, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(439, __Cat, [651 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(652, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 653, [652 | __Ss], [__T | __Stack]);
yeccpars2(652, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(654, __Cat, [652 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(653, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 436, [653 | __Ss], [__T | __Stack]);
yeccpars2(653, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [653 | __Ss], [__T | __Stack]);
yeccpars2(653, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(439, __Cat, [653 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(654, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 655, [654 | __Ss], [__T | __Stack]);
yeccpars2(654, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(655, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'ObservedEventsDescriptor'{requestId = __3, observedEventLst = [__5|__6]},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto(observedEventsDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(656, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 653, [656 | __Ss], [__T | __Stack]);
yeccpars2(656, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(657, __Cat, [656 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(657, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(observedEvents, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(658, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 660, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 659, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 661, [658 | __Ss], [__T | __Stack]);
yeccpars2(658, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(659, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  emergencyAudit,
 yeccpars2(yeccgoto(contextAuditProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(660, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  priorityAudit,
 yeccpars2(yeccgoto(contextAuditProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(661, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  topologyAudit,
 yeccpars2(yeccgoto(contextAuditProperty, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(662, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 663, [662 | __Ss], [__T | __Stack]);
yeccpars2(662, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(664, __Cat, [662 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(663, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 660, [663 | __Ss], [__T | __Stack]);
yeccpars2(663, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 659, [663 | __Ss], [__T | __Stack]);
yeccpars2(663, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 661, [663 | __Ss], [__T | __Stack]);
yeccpars2(663, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(664, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 665, [664 | __Ss], [__T | __Stack]);
yeccpars2(664, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(665, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(contextAudit, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(666, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 663, [666 | __Ss], [__T | __Stack]);
yeccpars2(666, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(667, __Cat, [666 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(667, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(contextAuditProperties, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(668, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [668 | __Ss], [__T | __Stack]);
yeccpars2(668, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(669, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 586, [669 | __Ss], [__T | __Stack]);
yeccpars2(669, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(670, __Cat, [669 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(670, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_commandRequest({auditValueRequest,__1},#'AuditRequest'{terminationID = __3, auditDescriptor = __4}),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(auditRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(671, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [671 | __Ss], [__T | __Stack]);
yeccpars2(671, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(672, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 586, [672 | __Ss], [__T | __Stack]);
yeccpars2(672, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(673, __Cat, [672 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(673, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_commandRequest({auditCapRequest,__1},#'AuditRequest'{terminationID = __3, auditDescriptor = __4}),
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(auditRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(674, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [674 | __Ss], [__T | __Stack]);
yeccpars2(674, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(675, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_uint32(__1),
 yeccpars2(yeccgoto(transactionID, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(676, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 677, [676 | __Ss], [__T | __Stack]);
yeccpars2(676, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(677, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 122, [677 | __Ss], [__T | __Stack]);
yeccpars2(677, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(678, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [678 | __Ss], [__T | __Stack]);
yeccpars2(678, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(679, __Cat, [678 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(679, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 680, [679 | __Ss], [__T | __Stack]);
yeccpars2(679, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(680, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'TransactionRequest'{transactionId = ensure_transactionID(__3), actions = [__5|__6]},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto(transactionRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(681, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 124, [681 | __Ss], [__T | __Stack]);
yeccpars2(681, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(682, __Cat, [681 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(682, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 683, [682 | __Ss], [__T | __Stack]);
yeccpars2(682, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(683, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'TransactionRequest'{transactionId = asn1_NOVALUE, actions = [__4|__5]},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(transactionRequest, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(684, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [684 | __Ss], [__T | __Stack]);
yeccpars2(684, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(685, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_transactionAck(__1),
 yeccpars2(yeccgoto(transactionAck, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(686, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 687, [686 | __Ss], [__T | __Stack]);
yeccpars2(686, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(688, __Cat, [686 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(687, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [687 | __Ss], [__T | __Stack]);
yeccpars2(687, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(688, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 689, [688 | __Ss], [__T | __Stack]);
yeccpars2(688, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(689, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(transactionResponseAck, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(690, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 687, [690 | __Ss], [__T | __Stack]);
yeccpars2(690, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(691, __Cat, [690 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(691, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(transactionAckList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(692, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [692 | __Ss], [__T | __Stack]);
yeccpars2(692, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(693, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 694, [693 | __Ss], [__T | __Stack]);
yeccpars2(693, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(694, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 695, [694 | __Ss], [__T | __Stack]);
yeccpars2(694, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(696, __Cat, [694 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(695, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 833, [695 | __Ss], [__T | __Stack]);
yeccpars2(695, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(696, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [696 | __Ss], [__T | __Stack]);
yeccpars2(696, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 697, [696 | __Ss], [__T | __Stack]);
yeccpars2(696, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(697, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 706, [697 | __Ss], [__T | __Stack]);
yeccpars2(697, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(698, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 702, [698 | __Ss], [__T | __Stack]);
yeccpars2(698, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(703, __Cat, [698 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(699, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {transactionError,__1},
 yeccpars2(yeccgoto(transactionReplyBody, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(700, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 701, [700 | __Ss], [__T | __Stack]);
yeccpars2(700, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(701, __Cat, __Ss,  [__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'TransactionReply'{transactionId = __3, immAckRequired = __5, transactionResult = __6},
 __Nss = lists:nthtail(6, __Ss),
 yeccpars2(yeccgoto(transactionReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(702, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 697, [702 | __Ss], [__T | __Stack]);
yeccpars2(702, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(703, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {actionReplies,[__1|__2]},
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(transactionReplyBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(704, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 702, [704 | __Ss], [__T | __Stack]);
yeccpars2(704, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(705, __Cat, [704 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(705, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(actionReplyList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(706, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [706 | __Ss], [__T | __Stack]);
yeccpars2(706, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(707, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 708, [707 | __Ss], [__T | __Stack]);
yeccpars2(707, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(708, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [708 | __Ss], [__T | __Stack]);
yeccpars2(708, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [708 | __Ss], [__T | __Stack]);
yeccpars2(708, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [708 | __Ss], [__T | __Stack]);
yeccpars2(708, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 715, [708 | __Ss], [__T | __Stack]);
yeccpars2(708, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 714, [708 | __Ss], [__T | __Stack]);
yeccpars2(708, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 710, [708 | __Ss], [__T | __Stack]);
yeccpars2(708, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 711, [708 | __Ss], [__T | __Stack]);
yeccpars2(708, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 716, [708 | __Ss], [__T | __Stack]);
yeccpars2(708, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 712, [708 | __Ss], [__T | __Stack]);
yeccpars2(708, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 713, [708 | __Ss], [__T | __Stack]);
yeccpars2(708, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 709, [708 | __Ss], [__T | __Stack]);
yeccpars2(708, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [708 | __Ss], [__T | __Stack]);
yeccpars2(708, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(709, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  addReply,
 yeccpars2(yeccgoto(ammsToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(710, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 829, [710 | __Ss], [__T | __Stack]);
yeccpars2(710, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(711, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 809, [711 | __Ss], [__T | __Stack]);
yeccpars2(711, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(712, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  modReply,
 yeccpars2(yeccgoto(ammsToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(713, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  moveReply,
 yeccpars2(yeccgoto(ammsToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(714, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 803, [714 | __Ss], [__T | __Stack]);
yeccpars2(714, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(715, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 778, [715 | __Ss], [__T | __Stack]);
yeccpars2(715, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(716, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  subtractReply,
 yeccpars2(yeccgoto(ammsToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(717, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 777, [717 | __Ss], [__T | __Stack]);
yeccpars2(717, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(718, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {command,__1},
 yeccpars2(yeccgoto(commandReply, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(719, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 731, [719 | __Ss], [__T | __Stack]);
yeccpars2(719, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(720, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {command,__1},
 yeccpars2(yeccgoto(commandReply, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(721, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 726, [721 | __Ss], [__T | __Stack]);
yeccpars2(721, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(727, __Cat, [721 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(722, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {context,__1},
 yeccpars2(yeccgoto(commandReply, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(723, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'ActionReply'{errorDescriptor = __1},
 yeccpars2(yeccgoto(actionReplyBody, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(724, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {command,__1},
 yeccpars2(yeccgoto(commandReply, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(725, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {command,__1},
 yeccpars2(yeccgoto(commandReply, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(726, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 141, [726 | __Ss], [__T | __Stack]);
yeccpars2(726, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 144, [726 | __Ss], [__T | __Stack]);
yeccpars2(726, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [726 | __Ss], [__T | __Stack]);
yeccpars2(726, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 715, [726 | __Ss], [__T | __Stack]);
yeccpars2(726, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 714, [726 | __Ss], [__T | __Stack]);
yeccpars2(726, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 710, [726 | __Ss], [__T | __Stack]);
yeccpars2(726, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 711, [726 | __Ss], [__T | __Stack]);
yeccpars2(726, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 716, [726 | __Ss], [__T | __Stack]);
yeccpars2(726, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 712, [726 | __Ss], [__T | __Stack]);
yeccpars2(726, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 713, [726 | __Ss], [__T | __Stack]);
yeccpars2(726, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 709, [726 | __Ss], [__T | __Stack]);
yeccpars2(726, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 137, [726 | __Ss], [__T | __Stack]);
yeccpars2(726, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(727, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_action_reply([__1|__2]),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(actionReplyBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(728, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 726, [728 | __Ss], [__T | __Stack]);
yeccpars2(728, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(730, __Cat, [728 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(729, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(commandReplyList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(730, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(commandReplyList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(731, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [731 | __Ss], [__T | __Stack]);
yeccpars2(731, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(732, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 733, [732 | __Ss], [__T | __Stack]);
yeccpars2(732, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(734, __Cat, [732 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(733, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 740, [733 | __Ss], [__T | __Stack]);
yeccpars2(733, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 739, [733 | __Ss], [__T | __Stack]);
yeccpars2(733, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 163, [733 | __Ss], [__T | __Stack]);
yeccpars2(733, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 736, [733 | __Ss], [__T | __Stack]);
yeccpars2(733, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 738, [733 | __Ss], [__T | __Stack]);
yeccpars2(733, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 169, [733 | __Ss], [__T | __Stack]);
yeccpars2(733, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [733 | __Ss], [__T | __Stack]);
yeccpars2(733, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 737, [733 | __Ss], [__T | __Stack]);
yeccpars2(733, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [733 | __Ss], [__T | __Stack]);
yeccpars2(733, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 735, [733 | __Ss], [__T | __Stack]);
yeccpars2(733, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [733 | __Ss], [__T | __Stack]);
yeccpars2(733, 'DigitMapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 463, [733 | __Ss], [__T | __Stack]);
yeccpars2(733, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(734, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {__1,#'AmmsReply'{terminationID = [__3], terminationAudit = __4}},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(ammsReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(735, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 301, [735 | __Ss], [__T | __Stack]);
yeccpars2(735, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  mediaToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(736, 'LSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 281, [736 | __Ss], [__T | __Stack]);
yeccpars2(736, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 280, [736 | __Ss], [__T | __Stack]);
yeccpars2(736, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  modemToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(737, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 269, [737 | __Ss], [__T | __Stack]);
yeccpars2(737, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  muxToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(738, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 649, [738 | __Ss], [__T | __Stack]);
yeccpars2(738, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  observedEventsToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(739, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 770, [739 | __Ss], [__T | __Stack]);
yeccpars2(739, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  packagesToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(740, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 760, [740 | __Ss], [__T | __Stack]);
yeccpars2(740, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  statsToken,
 yeccpars2(yeccgoto(auditReturnItem, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(741, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditReturnItem,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(742, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 756, [742 | __Ss], [__T | __Stack]);
yeccpars2(742, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(757, __Cat, [742 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(743, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {digitMapDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(744, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {errorDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(745, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {eventBufferDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(746, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {eventsDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(747, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {mediaDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(748, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {modemDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(749, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {muxDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(750, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {observedEventsDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(751, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {packagesDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(752, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {signalsDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(753, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {statisticsDescriptor,__1},
 yeccpars2(yeccgoto(auditReturnParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(754, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 755, [754 | __Ss], [__T | __Stack]);
yeccpars2(754, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(755, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(ammsReplyBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(756, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 740, [756 | __Ss], [__T | __Stack]);
yeccpars2(756, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 739, [756 | __Ss], [__T | __Stack]);
yeccpars2(756, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 163, [756 | __Ss], [__T | __Stack]);
yeccpars2(756, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 736, [756 | __Ss], [__T | __Stack]);
yeccpars2(756, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 738, [756 | __Ss], [__T | __Stack]);
yeccpars2(756, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 169, [756 | __Ss], [__T | __Stack]);
yeccpars2(756, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [756 | __Ss], [__T | __Stack]);
yeccpars2(756, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 737, [756 | __Ss], [__T | __Stack]);
yeccpars2(756, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [756 | __Ss], [__T | __Stack]);
yeccpars2(756, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 735, [756 | __Ss], [__T | __Stack]);
yeccpars2(756, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [756 | __Ss], [__T | __Stack]);
yeccpars2(756, 'DigitMapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 463, [756 | __Ss], [__T | __Stack]);
yeccpars2(756, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(757, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_terminationAudit([__1|__2]),
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(terminationAudit, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(758, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 756, [758 | __Ss], [__T | __Stack]);
yeccpars2(758, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(759, __Cat, [758 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(759, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(auditReturnParameterList, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(760, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [760 | __Ss], [__T | __Stack]);
yeccpars2(760, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(761, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 768, [761 | __Ss], [__T | __Stack]);
yeccpars2(761, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'StatisticsParameter'{statName = __1, statValue = asn1_NOVALUE},
 yeccpars2(yeccgoto(statisticsParameter, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(762, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 763, [762 | __Ss], [__T | __Stack]);
yeccpars2(762, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(764, __Cat, [762 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(763, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [763 | __Ss], [__T | __Stack]);
yeccpars2(763, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(764, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 765, [764 | __Ss], [__T | __Stack]);
yeccpars2(764, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(765, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(statisticsDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(766, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 763, [766 | __Ss], [__T | __Stack]);
yeccpars2(766, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(767, __Cat, [766 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(767, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(statisticsParameters, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(768, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 215, [768 | __Ss], [__T | __Stack]);
yeccpars2(768, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(769, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'StatisticsParameter'{statName = __1, statValue = [__3]},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(statisticsParameter, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(770, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [770 | __Ss], [__T | __Stack]);
yeccpars2(770, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(771, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 772, [771 | __Ss], [__T | __Stack]);
yeccpars2(771, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(773, __Cat, [771 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(772, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [772 | __Ss], [__T | __Stack]);
yeccpars2(772, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(773, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 774, [773 | __Ss], [__T | __Stack]);
yeccpars2(773, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(774, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__3|__4],
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(packagesDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(775, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 772, [775 | __Ss], [__T | __Stack]);
yeccpars2(775, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(776, __Cat, [775 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(776, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(packagesItems, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(777, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  setelement(#'ActionReply'.contextId,__5,__3),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(actionReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(778, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [778 | __Ss], [__T | __Stack]);
yeccpars2(778, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(779, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 780, [779 | __Ss], [__T | __Stack]);
yeccpars2(779, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  {serviceChangeResParms,#'ServiceChangeResParm'{}},
 yeccpars2(781, __Cat, [779 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(780, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 782, [780 | __Ss], [__T | __Stack]);
yeccpars2(780, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [780 | __Ss], [__T | __Stack]);
yeccpars2(780, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(781, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {serviceChangeReply,#'ServiceChangeReply'{terminationID = [__3], serviceChangeResult = __4}},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(serviceChangeReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(782, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 787, [782 | __Ss], [__T | __Stack]);
yeccpars2(782, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(783, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 786, [783 | __Ss], [__T | __Stack]);
yeccpars2(783, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(784, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 785, [784 | __Ss], [__T | __Stack]);
yeccpars2(784, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(785, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {serviceChangeResParms,__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeReplyBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(786, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {errorDescriptor,__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(serviceChangeReplyBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(787, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 436, [787 | __Ss], [__T | __Stack]);
yeccpars2(787, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 791, [787 | __Ss], [__T | __Stack]);
yeccpars2(787, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 789, [787 | __Ss], [__T | __Stack]);
yeccpars2(787, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 788, [787 | __Ss], [__T | __Stack]);
yeccpars2(787, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 790, [787 | __Ss], [__T | __Stack]);
yeccpars2(787, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(788, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 633, [788 | __Ss], [__T | __Stack]);
yeccpars2(788, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(789, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 631, [789 | __Ss], [__T | __Stack]);
yeccpars2(789, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(790, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 625, [790 | __Ss], [__T | __Stack]);
yeccpars2(790, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(791, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 623, [791 | __Ss], [__T | __Stack]);
yeccpars2(791, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(792, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 798, [792 | __Ss], [__T | __Stack]);
yeccpars2(792, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(799, __Cat, [792 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(793, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {address,__1},
 yeccpars2(yeccgoto(servChgReplyParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(794, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {mgc_id,__1},
 yeccpars2(yeccgoto(servChgReplyParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(795, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {profile,__1},
 yeccpars2(yeccgoto(servChgReplyParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(796, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {version,__1},
 yeccpars2(yeccgoto(servChgReplyParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(797, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {time_stamp,__1},
 yeccpars2(yeccgoto(servChgReplyParm, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(798, 'TimeStampToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 436, [798 | __Ss], [__T | __Stack]);
yeccpars2(798, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 791, [798 | __Ss], [__T | __Stack]);
yeccpars2(798, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 789, [798 | __Ss], [__T | __Stack]);
yeccpars2(798, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 788, [798 | __Ss], [__T | __Stack]);
yeccpars2(798, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 790, [798 | __Ss], [__T | __Stack]);
yeccpars2(798, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(799, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 800, [799 | __Ss], [__T | __Stack]);
yeccpars2(799, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(800, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  merge_ServiceChangeResParm([__3|__4]),
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(serviceChangeReplyDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(801, 'COMMA', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 798, [801 | __Ss], [__T | __Stack]);
yeccpars2(801, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(802, __Cat, [801 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(802, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__2|__3],
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(servChgReplyParms, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(803, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [803 | __Ss], [__T | __Stack]);
yeccpars2(803, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(804, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 805, [804 | __Ss], [__T | __Stack]);
yeccpars2(804, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(806, __Cat, [804 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(805, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [805 | __Ss], [__T | __Stack]);
yeccpars2(805, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(806, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {notifyReply,#'NotifyReply'{terminationID = [__3], errorDescriptor = __4}},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(notifyReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(807, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 808, [807 | __Ss], [__T | __Stack]);
yeccpars2(807, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(808, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  __2,
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(notifyReplyBody, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(809, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 810, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [809 | __Ss], [__T | __Stack]);
yeccpars2(809, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(810, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 816, [810 | __Ss], [__T | __Stack]);
yeccpars2(810, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(811, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditValueReply,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(auditReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(812, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 813, [812 | __Ss], [__T | __Stack]);
yeccpars2(812, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditResult,#'AuditResult'{terminationID = __1, terminationAuditResult = []}},
 yeccpars2(yeccgoto(auditOther, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(813, 'StatsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 740, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'PackagesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 739, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'DigitMapDescriptorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 163, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'ModemToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 736, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'ObservedEventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 738, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'SignalsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 169, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'EventsToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 165, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'MuxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 737, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'EventBufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 164, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'MediaToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 735, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 106, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, 'DigitMapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 463, [813 | __Ss], [__T | __Stack]);
yeccpars2(813, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(814, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 815, [814 | __Ss], [__T | __Stack]);
yeccpars2(814, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(815, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditResult,#'AuditResult'{terminationID = __1, terminationAuditResult = __3}},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(auditOther, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(816, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 819, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [816 | __Ss], [__T | __Stack]);
yeccpars2(816, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(817, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditValueReply,__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(auditReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(818, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {contextAuditResult,__1},
 yeccpars2(yeccgoto(contextTerminationAudit, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(819, 'EQUAL', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 822, [819 | __Ss], [__T | __Stack]);
yeccpars2(819, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(820, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 821, [820 | __Ss], [__T | __Stack]);
yeccpars2(820, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(821, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {contextAuditResult,__2},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(contextTerminationAudit, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(822, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [822 | __Ss], [__T | __Stack]);
yeccpars2(822, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(823, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 825, [823 | __Ss], [__T | __Stack]);
yeccpars2(823, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(824, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_uint(__1,0,999),
 yeccpars2(yeccgoto(errorCode, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(825, 'QuotedChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 826, [825 | __Ss], [__T | __Stack]);
yeccpars2(825, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  asn1_NOVALUE,
 yeccpars2(827, __Cat, [825 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(826, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  value_of(__1),
 yeccpars2(yeccgoto(errorText, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(827, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 828, [827 | __Ss], [__T | __Stack]);
yeccpars2(827, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(828, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'ErrorDescriptor'{errorCode = __3, errorText = __5},
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(errorDescriptor, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(829, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 830, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [829 | __Ss], [__T | __Stack]);
yeccpars2(829, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(830, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 816, [830 | __Ss], [__T | __Stack]);
yeccpars2(830, __Cat, __Ss,  [__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  make_safe_token(__1),
 yeccpars2(yeccgoto(safeToken, hd(__Ss)), __Cat, __Ss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(831, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditCapReply,__3},
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(auditReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(832, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  {auditCapReply,__4},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(auditReply, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(833, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  'NULL',
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(optImmAckRequired, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(834, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [834 | __Ss], [__T | __Stack]);
yeccpars2(834, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(835, 'LBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 836, [835 | __Ss], [__T | __Stack]);
yeccpars2(835, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(836, 'RBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 837, [836 | __Ss], [__T | __Stack]);
yeccpars2(836, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(837, __Cat, __Ss,  [__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'TransactionPending'{transactionId = ensure_transactionID(__3)},
 __Nss = lists:nthtail(4, __Ss),
 yeccpars2(yeccgoto(transactionPending, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(838, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 838, [838 | __Ss], [__T | __Stack]);
yeccpars2(838, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(846, __Cat, [838 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(839, 'RSBRKT', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 842, [839 | __Ss], [__T | __Stack]);
yeccpars2(839, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(840, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 838, [840 | __Ss], [__T | __Stack]);
yeccpars2(840, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  [],
 yeccpars2(841, __Cat, [840 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(841, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [__1|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(daddr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(842, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 843, [842 | __Ss], [__T | __Stack]);
yeccpars2(842, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_domainAddress(__2,asn1_NOVALUE),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(domainAddress, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(843, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [843 | __Ss], [__T | __Stack]);
yeccpars2(843, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(844, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [844 | __Ss], [__T | __Stack]);
yeccpars2(844, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(845, __Cat, [844 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(845, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_domainAddress(__2,__5),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(domainAddress, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(846, __Cat, __Ss,  [__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  [colon|__2],
 __Nss = lists:nthtail(1, __Ss),
 yeccpars2(yeccgoto(daddr, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(847, 'GREATER', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 848, [847 | __Ss], [__T | __Stack]);
yeccpars2(847, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(848, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 849, [848 | __Ss], [__T | __Stack]);
yeccpars2(848, __Cat, __Ss,  [__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_domainName(__2,asn1_NOVALUE),
 __Nss = lists:nthtail(2, __Ss),
 yeccpars2(yeccgoto(domainName, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(849, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [849 | __Ss], [__T | __Stack]);
yeccpars2(849, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(850, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [850 | __Ss], [__T | __Stack]);
yeccpars2(850, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(851, __Cat, [850 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(851, __Cat, __Ss,  [__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_domainName(__2,__5),
 __Nss = lists:nthtail(5, __Ss),
 yeccpars2(yeccgoto(domainName, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(852, __Cat, __Ss,  [__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  #'MegacoMessage'{authHeader = __2, mess = __3},
 __Nss = lists:nthtail(3, __Ss),
 yeccpars2(yeccgoto(megacoMessage, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(853, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [853 | __Ss], [__T | __Stack]);
yeccpars2(853, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(854, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 855, [854 | __Ss], [__T | __Stack]);
yeccpars2(854, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(855, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [855 | __Ss], [__T | __Stack]);
yeccpars2(855, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(856, 'COLON', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 857, [856 | __Ss], [__T | __Stack]);
yeccpars2(856, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(857, 'VersionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 90, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'V91Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 89, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'V90Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 88, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'V76Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 87, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'V34Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 86, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'V32bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 85, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'V32Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 84, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'V22bisToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 83, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'V22Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 82, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'V18Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 81, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'TransToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 80, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'TopologyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 79, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'TimeOutToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 78, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'TestToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 77, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'TerminationStateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 76, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'SynchISDNToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 75, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'SubtractToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 74, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'StreamToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 73, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'SignalTypeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 72, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'SignalListToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 71, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ServiceChangeAddressToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 67, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ServiceChangeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 68, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ServiceStatesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 69, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ServicesToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 70, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'SendrecvToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 66, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'SendonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 65, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ReservedValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 61, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ReservedGroupToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 60, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'RestartToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 63, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ResponseAckToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 62, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ReplyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 59, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'RecvonlyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 58, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 57, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ProfileToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 56, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'PriorityToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 55, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'PendingToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 54, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'OtherReasonToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 52, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'OutOfSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 53, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'OnOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 49, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'OnToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 50, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'OffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 48, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'OnewayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 51, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'Nx64Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 47, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'NotifyCompletionToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 45, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'NotifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 46, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'MoveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 44, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ModifyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 43, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ModeToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 42, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'MgcIdToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 41, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'MethodToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 40, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'LockStepToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 38, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'LoopbackToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 39, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'LocalControlToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 37, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'KeepActiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 36, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'InSvcToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 31, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'IsolateToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 35, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'InterruptByNewSignalsDescrToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 34, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'InterruptByEventToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 33, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'InactiveToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 32, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ImmAckRequiredToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 30, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'HandOffToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 29, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'H226Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 28, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'H223Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 27, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'H221Token', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 26, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'GracefulToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 25, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ForcedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 24, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'FailoverToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 23, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ErrorToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 22, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'EmergencyToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 21, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'EmbedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 20, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'DurationToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 19, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'DelayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 16, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'DisconnectedToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 18, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'DiscardToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 17, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'ContextAuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 14, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'CtxToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 15, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'BufferToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 13, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'BriefToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 12, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'BothwayToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 11, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'AuthToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 10, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'AuditValueToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 9, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'AuditCapToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 7, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'AuditToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 8, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'AddToken', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 6, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, 'SafeChars', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 64, [857 | __Ss], [__T | __Stack]);
yeccpars2(857, _, _, _, __T, _, _) ->
 yeccerror(__T);
yeccpars2(858, 'SEP', __Ss, __Stack, __T, __Ts, __Tzr) ->
 yeccpars1(__Ts, __Tzr, 1, [858 | __Ss], [__T | __Stack]);
yeccpars2(858, __Cat, __Ss,  __Stack, __T, __Ts, __Tzr) ->
 __Val =  no_sep,
 yeccpars2(859, __Cat, [858 | __Ss], [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(859, __Cat, __Ss,  [__8,__7,__6,__5,__4,__3,__2,__1|__Stack], __T, __Ts, __Tzr) ->
 __Val =  ensure_auth_header(__3,__5,__7),
 __Nss = lists:nthtail(7, __Ss),
 yeccpars2(yeccgoto(authenticationHeader, hd(__Nss)), __Cat, __Nss, [__Val | __Stack], __T, __Ts, __Tzr);
yeccpars2(__Other, _, _, _, _, _, _) ->
 exit({parser, __Other, missing_state_in_action_table}).

yeccgoto(actionReply, 696) ->
 698;
yeccgoto(actionReply, 702) ->
 704;
yeccgoto(actionReplyBody, 708) ->
 717;
yeccgoto(actionReplyList, 698) ->
 703;
yeccgoto(actionReplyList, 704) ->
 705;
yeccgoto(actionRequest, 121) ->
 123;
yeccgoto(actionRequest, 124) ->
 127;
yeccgoto(actionRequest, 674) ->
 681;
yeccgoto(actionRequest, 677) ->
 678;
yeccgoto(actionRequestItem, 132) ->
 145;
yeccgoto(actionRequestItem, 562) ->
 565;
yeccgoto(actionRequestItems, 145) ->
 563;
yeccgoto(actionRequestItems, 565) ->
 566;
yeccgoto(actionRequestList, 123) ->
 125;
yeccgoto(actionRequestList, 127) ->
 128;
yeccgoto(actionRequestList, 678) ->
 679;
yeccgoto(actionRequestList, 681) ->
 682;
yeccgoto(alternativeValue, 210) ->
 222;
yeccgoto(ammParameter, 160) ->
 170;
yeccgoto(ammParameter, 179) ->
 182;
yeccgoto(ammParameters, 170) ->
 180;
yeccgoto(ammParameters, 182) ->
 183;
yeccgoto(ammRequest, 132) ->
 146;
yeccgoto(ammRequest, 562) ->
 146;
yeccgoto(ammRequestBody, 159) ->
 161;
yeccgoto(ammToken, 132) ->
 147;
yeccgoto(ammToken, 562) ->
 147;
yeccgoto(ammsReply, 708) ->
 718;
yeccgoto(ammsReply, 726) ->
 718;
yeccgoto(ammsReplyBody, 732) ->
 734;
yeccgoto(ammsToken, 708) ->
 719;
yeccgoto(ammsToken, 726) ->
 719;
yeccgoto(auditDescriptor, 160) ->
 171;
yeccgoto(auditDescriptor, 179) ->
 171;
yeccgoto(auditDescriptor, 586) ->
 588;
yeccgoto(auditDescriptorBody, 461) ->
 473;
yeccgoto(auditItem, 461) ->
 474;
yeccgoto(auditItem, 557) ->
 559;
yeccgoto(auditItem, 596) ->
 604;
yeccgoto(auditItem, 617) ->
 604;
yeccgoto(auditItemList, 474) ->
 558;
yeccgoto(auditItemList, 559) ->
 560;
yeccgoto(auditOther, 809) ->
 811;
yeccgoto(auditOther, 829) ->
 831;
yeccgoto(auditReply, 708) ->
 720;
yeccgoto(auditReply, 726) ->
 720;
yeccgoto(auditRequest, 132) ->
 148;
yeccgoto(auditRequest, 562) ->
 148;
yeccgoto(auditReturnItem, 461) ->
 475;
yeccgoto(auditReturnItem, 557) ->
 475;
yeccgoto(auditReturnItem, 596) ->
 475;
yeccgoto(auditReturnItem, 617) ->
 475;
yeccgoto(auditReturnItem, 733) ->
 741;
yeccgoto(auditReturnItem, 756) ->
 741;
yeccgoto(auditReturnItem, 813) ->
 741;
yeccgoto(auditReturnParameter, 733) ->
 742;
yeccgoto(auditReturnParameter, 756) ->
 758;
yeccgoto(auditReturnParameter, 813) ->
 742;
yeccgoto(auditReturnParameterList, 742) ->
 757;
yeccgoto(auditReturnParameterList, 758) ->
 759;
yeccgoto(authenticationHeader, 3) ->
 5;
yeccgoto(commandReply, 708) ->
 721;
yeccgoto(commandReply, 726) ->
 728;
yeccgoto(commandReplyList, 721) ->
 727;
yeccgoto(commandReplyList, 728) ->
 730;
yeccgoto(commandRequest, 132) ->
 149;
yeccgoto(commandRequest, 562) ->
 149;
yeccgoto(contextAudit, 132) ->
 150;
yeccgoto(contextAudit, 562) ->
 150;
yeccgoto(contextAuditProperties, 662) ->
 664;
yeccgoto(contextAuditProperties, 666) ->
 667;
yeccgoto(contextAuditProperty, 658) ->
 662;
yeccgoto(contextAuditProperty, 663) ->
 666;
yeccgoto(contextID, 129) ->
 130;
yeccgoto(contextID, 706) ->
 707;
yeccgoto(contextProperty, 132) ->
 151;
yeccgoto(contextProperty, 562) ->
 151;
yeccgoto(contextProperty, 708) ->
 722;
yeccgoto(contextProperty, 726) ->
 722;
yeccgoto(contextTerminationAudit, 810) ->
 817;
yeccgoto(contextTerminationAudit, 830) ->
 832;
yeccgoto(daddr, 94) ->
 839;
yeccgoto(daddr, 838) ->
 846;
yeccgoto(daddr, 840) ->
 841;
yeccgoto(deviceName, 98) ->
 100;
yeccgoto(digitMapDescriptor, 160) ->
 172;
yeccgoto(digitMapDescriptor, 179) ->
 172;
yeccgoto(digitMapDescriptor, 733) ->
 743;
yeccgoto(digitMapDescriptor, 756) ->
 743;
yeccgoto(digitMapDescriptor, 813) ->
 743;
yeccgoto(domainAddress, 92) ->
 95;
yeccgoto(domainAddress, 625) ->
 95;
yeccgoto(domainAddress, 633) ->
 95;
yeccgoto(domainName, 92) ->
 96;
yeccgoto(domainName, 625) ->
 96;
yeccgoto(domainName, 633) ->
 96;
yeccgoto(embedFirst, 400) ->
 402;
yeccgoto(embedFirst, 404) ->
 406;
yeccgoto(embedNoSig, 382) ->
 387;
yeccgoto(embedNoSig, 395) ->
 387;
yeccgoto(embedSig, 419) ->
 423;
yeccgoto(embedSig, 427) ->
 423;
yeccgoto(embedWithSig, 382) ->
 388;
yeccgoto(embedWithSig, 395) ->
 388;
yeccgoto(errorCode, 822) ->
 823;
yeccgoto(errorDescriptor, 97) ->
 111;
yeccgoto(errorDescriptor, 643) ->
 645;
yeccgoto(errorDescriptor, 696) ->
 699;
yeccgoto(errorDescriptor, 708) ->
 723;
yeccgoto(errorDescriptor, 726) ->
 729;
yeccgoto(errorDescriptor, 733) ->
 744;
yeccgoto(errorDescriptor, 756) ->
 744;
yeccgoto(errorDescriptor, 780) ->
 783;
yeccgoto(errorDescriptor, 805) ->
 807;
yeccgoto(errorDescriptor, 813) ->
 744;
yeccgoto(errorDescriptor, 816) ->
 820;
yeccgoto(errorText, 825) ->
 827;
yeccgoto(eventBufferControl, 317) ->
 320;
yeccgoto(eventBufferControl, 324) ->
 320;
yeccgoto(eventBufferControlState, 334) ->
 337;
yeccgoto(eventBufferDescriptor, 160) ->
 173;
yeccgoto(eventBufferDescriptor, 179) ->
 173;
yeccgoto(eventBufferDescriptor, 733) ->
 745;
yeccgoto(eventBufferDescriptor, 756) ->
 745;
yeccgoto(eventBufferDescriptor, 813) ->
 745;
yeccgoto(eventDM, 382) ->
 389;
yeccgoto(eventDM, 395) ->
 389;
yeccgoto(eventDM, 419) ->
 424;
yeccgoto(eventDM, 427) ->
 424;
yeccgoto(eventParameter, 382) ->
 390;
yeccgoto(eventParameter, 395) ->
 398;
yeccgoto(eventParameterName, 382) ->
 391;
yeccgoto(eventParameterName, 395) ->
 391;
yeccgoto(eventParameterName, 419) ->
 391;
yeccgoto(eventParameterName, 427) ->
 391;
yeccgoto(eventParameterName, 445) ->
 391;
yeccgoto(eventParameterName, 449) ->
 391;
yeccgoto(eventParameterName, 547) ->
 550;
yeccgoto(eventParameters, 390) ->
 396;
yeccgoto(eventParameters, 398) ->
 399;
yeccgoto(eventSpec, 435) ->
 437;
yeccgoto(eventSpec, 456) ->
 459;
yeccgoto(eventSpecList, 437) ->
 457;
yeccgoto(eventSpecList, 459) ->
 460;
yeccgoto(eventStream, 547) ->
 551;
yeccgoto(eventStreamOrOther, 382) ->
 392;
yeccgoto(eventStreamOrOther, 395) ->
 392;
yeccgoto(eventStreamOrOther, 419) ->
 425;
yeccgoto(eventStreamOrOther, 427) ->
 425;
yeccgoto(eventStreamOrOther, 445) ->
 447;
yeccgoto(eventStreamOrOther, 449) ->
 447;
yeccgoto(eventsDescriptor, 160) ->
 174;
yeccgoto(eventsDescriptor, 179) ->
 174;
yeccgoto(eventsDescriptor, 733) ->
 746;
yeccgoto(eventsDescriptor, 756) ->
 746;
yeccgoto(eventsDescriptor, 813) ->
 746;
yeccgoto(extension, 596) ->
 605;
yeccgoto(extension, 617) ->
 605;
yeccgoto(extensionParameter, 596) ->
 606;
yeccgoto(extensionParameter, 617) ->
 606;
yeccgoto(indAudauditReturnParameter, 461) ->
 476;
yeccgoto(indAudauditReturnParameter, 485) ->
 493;
yeccgoto(indAudauditReturnParameter, 557) ->
 476;
yeccgoto(indAudauditReturnParameter, 596) ->
 476;
yeccgoto(indAudauditReturnParameter, 617) ->
 476;
yeccgoto(indAuddigitMapDescriptor, 461) ->
 477;
yeccgoto(indAuddigitMapDescriptor, 485) ->
 477;
yeccgoto(indAuddigitMapDescriptor, 557) ->
 477;
yeccgoto(indAuddigitMapDescriptor, 596) ->
 477;
yeccgoto(indAuddigitMapDescriptor, 617) ->
 477;
yeccgoto(indAudeventBufferDescriptor, 461) ->
 478;
yeccgoto(indAudeventBufferDescriptor, 485) ->
 478;
yeccgoto(indAudeventBufferDescriptor, 557) ->
 478;
yeccgoto(indAudeventBufferDescriptor, 596) ->
 478;
yeccgoto(indAudeventBufferDescriptor, 617) ->
 478;
yeccgoto(indAudeventSpec, 544) ->
 545;
yeccgoto(indAudeventSpecParameter, 547) ->
 552;
yeccgoto(indAudeventsDescriptor, 461) ->
 479;
yeccgoto(indAudeventsDescriptor, 485) ->
 479;
yeccgoto(indAudeventsDescriptor, 557) ->
 479;
yeccgoto(indAudeventsDescriptor, 596) ->
 479;
yeccgoto(indAudeventsDescriptor, 617) ->
 479;
yeccgoto(indAudlocalControlDescriptor, 515) ->
 519;
yeccgoto(indAudlocalControlDescriptor, 531) ->
 519;
yeccgoto(indAudlocalParm, 534) ->
 535;
yeccgoto(indAudmediaDescriptor, 461) ->
 480;
yeccgoto(indAudmediaDescriptor, 485) ->
 480;
yeccgoto(indAudmediaDescriptor, 557) ->
 480;
yeccgoto(indAudmediaDescriptor, 596) ->
 480;
yeccgoto(indAudmediaDescriptor, 617) ->
 480;
yeccgoto(indAudmediaParm, 515) ->
 520;
yeccgoto(indAudpackagesDescriptor, 461) ->
 481;
yeccgoto(indAudpackagesDescriptor, 485) ->
 481;
yeccgoto(indAudpackagesDescriptor, 557) ->
 481;
yeccgoto(indAudpackagesDescriptor, 596) ->
 481;
yeccgoto(indAudpackagesDescriptor, 617) ->
 481;
yeccgoto(indAudrequestedEvent, 540) ->
 541;
yeccgoto(indAudsignalList, 498) ->
 502;
yeccgoto(indAudsignalParm, 498) ->
 503;
yeccgoto(indAudsignalsDescriptor, 461) ->
 482;
yeccgoto(indAudsignalsDescriptor, 485) ->
 482;
yeccgoto(indAudsignalsDescriptor, 557) ->
 482;
yeccgoto(indAudsignalsDescriptor, 596) ->
 482;
yeccgoto(indAudsignalsDescriptor, 617) ->
 482;
yeccgoto(indAudstatisticsDescriptor, 461) ->
 483;
yeccgoto(indAudstatisticsDescriptor, 485) ->
 483;
yeccgoto(indAudstatisticsDescriptor, 557) ->
 483;
yeccgoto(indAudstatisticsDescriptor, 596) ->
 483;
yeccgoto(indAudstatisticsDescriptor, 617) ->
 483;
yeccgoto(indAudstreamDescriptor, 515) ->
 521;
yeccgoto(indAudstreamParm, 515) ->
 522;
yeccgoto(indAudstreamParm, 531) ->
 532;
yeccgoto(indAudterminationAudit, 461) ->
 484;
yeccgoto(indAudterminationAudit, 557) ->
 484;
yeccgoto(indAudterminationAudit, 596) ->
 484;
yeccgoto(indAudterminationAudit, 617) ->
 484;
yeccgoto(indAudterminationAuditList, 476) ->
 486;
yeccgoto(indAudterminationAuditList, 493) ->
 494;
yeccgoto(indAudterminationStateDescriptor, 515) ->
 523;
yeccgoto(indAudterminationStateParm, 525) ->
 526;
yeccgoto(localControlDescriptor, 301) ->
 307;
yeccgoto(localControlDescriptor, 312) ->
 307;
yeccgoto(localControlDescriptor, 340) ->
 307;
yeccgoto(localControlDescriptor, 342) ->
 307;
yeccgoto(localParm, 347) ->
 351;
yeccgoto(localParm, 353) ->
 356;
yeccgoto(localParmList, 351) ->
 354;
yeccgoto(localParmList, 356) ->
 357;
yeccgoto(mId, 92) ->
 97;
yeccgoto(mId, 625) ->
 626;
yeccgoto(mId, 633) ->
 634;
yeccgoto(mediaDescriptor, 160) ->
 175;
yeccgoto(mediaDescriptor, 179) ->
 175;
yeccgoto(mediaDescriptor, 733) ->
 747;
yeccgoto(mediaDescriptor, 756) ->
 747;
yeccgoto(mediaDescriptor, 813) ->
 747;
yeccgoto(mediaParm, 301) ->
 308;
yeccgoto(mediaParm, 312) ->
 315;
yeccgoto(mediaParmList, 308) ->
 313;
yeccgoto(mediaParmList, 315) ->
 316;
yeccgoto(megacoMessage, 0) ->
 2;
yeccgoto(message, 5) ->
 91;
yeccgoto(messageBody, 97) ->
 112;
yeccgoto(modemDescriptor, 160) ->
 176;
yeccgoto(modemDescriptor, 179) ->
 176;
yeccgoto(modemDescriptor, 733) ->
 748;
yeccgoto(modemDescriptor, 756) ->
 748;
yeccgoto(modemDescriptor, 813) ->
 748;
yeccgoto(modemType, 280) ->
 299;
yeccgoto(modemType, 281) ->
 282;
yeccgoto(modemType, 284) ->
 297;
yeccgoto(modemTypeList, 282) ->
 285;
yeccgoto(modemTypeList, 297) ->
 298;
yeccgoto(mtpAddress, 98) ->
 101;
yeccgoto(muxDescriptor, 160) ->
 177;
yeccgoto(muxDescriptor, 179) ->
 177;
yeccgoto(muxDescriptor, 733) ->
 749;
yeccgoto(muxDescriptor, 756) ->
 749;
yeccgoto(muxDescriptor, 813) ->
 749;
yeccgoto(muxType, 269) ->
 270;
yeccgoto(notificationReason, 245) ->
 250;
yeccgoto(notificationReason, 251) ->
 254;
yeccgoto(notificationReasons, 250) ->
 252;
yeccgoto(notificationReasons, 254) ->
 255;
yeccgoto(notifyReply, 708) ->
 724;
yeccgoto(notifyReply, 726) ->
 724;
yeccgoto(notifyReplyBody, 804) ->
 806;
yeccgoto(notifyRequest, 132) ->
 152;
yeccgoto(notifyRequest, 562) ->
 152;
yeccgoto(notifyRequestBody, 643) ->
 646;
yeccgoto(observedEvent, 435) ->
 438;
yeccgoto(observedEvent, 456) ->
 438;
yeccgoto(observedEvent, 651) ->
 652;
yeccgoto(observedEvent, 653) ->
 656;
yeccgoto(observedEventBody, 444) ->
 446;
yeccgoto(observedEventBody, 454) ->
 455;
yeccgoto(observedEventParameter, 445) ->
 448;
yeccgoto(observedEventParameter, 449) ->
 452;
yeccgoto(observedEventParameters, 448) ->
 450;
yeccgoto(observedEventParameters, 452) ->
 453;
yeccgoto(observedEvents, 652) ->
 654;
yeccgoto(observedEvents, 656) ->
 657;
yeccgoto(observedEventsDescriptor, 643) ->
 647;
yeccgoto(observedEventsDescriptor, 733) ->
 750;
yeccgoto(observedEventsDescriptor, 756) ->
 750;
yeccgoto(observedEventsDescriptor, 813) ->
 750;
yeccgoto(onOrOff, 358) ->
 361;
yeccgoto(onOrOff, 362) ->
 363;
yeccgoto(optAuditDescriptor, 585) ->
 587;
yeccgoto(optAuditDescriptor, 669) ->
 670;
yeccgoto(optAuditDescriptor, 672) ->
 673;
yeccgoto(optImmAckRequired, 694) ->
 696;
yeccgoto(optIndAudeventSpecParameter, 546) ->
 548;
yeccgoto(optIndAudsignalParm, 471) ->
 499;
yeccgoto(optIndAudsignalParm, 491) ->
 499;
yeccgoto(optPropertyParms, 286) ->
 288;
yeccgoto(optPropertyParms, 299) ->
 300;
yeccgoto(optSep, 0) ->
 3;
yeccgoto(optSep, 92) ->
 98;
yeccgoto(optSep, 100) ->
 105;
yeccgoto(optSep, 101) ->
 104;
yeccgoto(optSep, 435) ->
 439;
yeccgoto(optSep, 440) ->
 441;
yeccgoto(optSep, 442) ->
 443;
yeccgoto(optSep, 456) ->
 439;
yeccgoto(optSep, 625) ->
 98;
yeccgoto(optSep, 633) ->
 98;
yeccgoto(optSep, 651) ->
 439;
yeccgoto(optSep, 653) ->
 439;
yeccgoto(optSep, 844) ->
 845;
yeccgoto(optSep, 850) ->
 851;
yeccgoto(optSep, 858) ->
 859;
yeccgoto(packagesDescriptor, 733) ->
 751;
yeccgoto(packagesDescriptor, 756) ->
 751;
yeccgoto(packagesDescriptor, 813) ->
 751;
yeccgoto(packagesItem, 511) ->
 512;
yeccgoto(packagesItem, 770) ->
 771;
yeccgoto(packagesItem, 772) ->
 775;
yeccgoto(packagesItems, 771) ->
 773;
yeccgoto(packagesItems, 775) ->
 776;
yeccgoto(parmValue, 203) ->
 214;
yeccgoto(parmValue, 289) ->
 296;
yeccgoto(parmValue, 391) ->
 394;
yeccgoto(parmValue, 606) ->
 622;
yeccgoto(pathName, 98) ->
 102;
yeccgoto(pkgdName, 184) ->
 186;
yeccgoto(pkgdName, 192) ->
 186;
yeccgoto(pkgdName, 261) ->
 186;
yeccgoto(pkgdName, 264) ->
 186;
yeccgoto(pkgdName, 287) ->
 289;
yeccgoto(pkgdName, 291) ->
 289;
yeccgoto(pkgdName, 317) ->
 289;
yeccgoto(pkgdName, 324) ->
 289;
yeccgoto(pkgdName, 347) ->
 289;
yeccgoto(pkgdName, 353) ->
 289;
yeccgoto(pkgdName, 374) ->
 375;
yeccgoto(pkgdName, 377) ->
 375;
yeccgoto(pkgdName, 411) ->
 412;
yeccgoto(pkgdName, 414) ->
 412;
yeccgoto(pkgdName, 439) ->
 454;
yeccgoto(pkgdName, 443) ->
 444;
yeccgoto(pkgdName, 495) ->
 496;
yeccgoto(pkgdName, 498) ->
 186;
yeccgoto(pkgdName, 508) ->
 186;
yeccgoto(pkgdName, 540) ->
 542;
yeccgoto(pkgdName, 544) ->
 546;
yeccgoto(pkgdName, 760) ->
 761;
yeccgoto(pkgdName, 763) ->
 761;
yeccgoto(portNumber, 625) ->
 627;
yeccgoto(portNumber, 843) ->
 844;
yeccgoto(portNumber, 849) ->
 850;
yeccgoto(priority, 132) ->
 153;
yeccgoto(priority, 562) ->
 153;
yeccgoto(priority, 708) ->
 153;
yeccgoto(priority, 726) ->
 153;
yeccgoto(propertyParm, 287) ->
 290;
yeccgoto(propertyParm, 291) ->
 294;
yeccgoto(propertyParm, 317) ->
 321;
yeccgoto(propertyParm, 324) ->
 321;
yeccgoto(propertyParm, 347) ->
 352;
yeccgoto(propertyParm, 353) ->
 352;
yeccgoto(propertyParms, 290) ->
 292;
yeccgoto(propertyParms, 294) ->
 295;
yeccgoto(requestID, 371) ->
 372;
yeccgoto(requestID, 409) ->
 410;
yeccgoto(requestID, 538) ->
 539;
yeccgoto(requestID, 649) ->
 650;
yeccgoto(requestedEvent, 374) ->
 376;
yeccgoto(requestedEvent, 377) ->
 380;
yeccgoto(requestedEventBody, 375) ->
 383;
yeccgoto(requestedEvents, 376) ->
 378;
yeccgoto(requestedEvents, 380) ->
 381;
yeccgoto(safeToken, 5) ->
 92;
yeccgoto(safeToken, 93) ->
 847;
yeccgoto(safeToken, 94) ->
 840;
yeccgoto(safeToken, 98) ->
 103;
yeccgoto(safeToken, 120) ->
 675;
yeccgoto(safeToken, 129) ->
 131;
yeccgoto(safeToken, 157) ->
 158;
yeccgoto(safeToken, 184) ->
 187;
yeccgoto(safeToken, 192) ->
 187;
yeccgoto(safeToken, 197) ->
 203;
yeccgoto(safeToken, 205) ->
 203;
yeccgoto(safeToken, 210) ->
 216;
yeccgoto(safeToken, 211) ->
 216;
yeccgoto(safeToken, 212) ->
 216;
yeccgoto(safeToken, 213) ->
 216;
yeccgoto(safeToken, 220) ->
 216;
yeccgoto(safeToken, 221) ->
 216;
yeccgoto(safeToken, 225) ->
 216;
yeccgoto(safeToken, 226) ->
 216;
yeccgoto(safeToken, 236) ->
 237;
yeccgoto(safeToken, 256) ->
 257;
yeccgoto(safeToken, 258) ->
 259;
yeccgoto(safeToken, 261) ->
 187;
yeccgoto(safeToken, 264) ->
 187;
yeccgoto(safeToken, 269) ->
 271;
yeccgoto(safeToken, 272) ->
 158;
yeccgoto(safeToken, 275) ->
 158;
yeccgoto(safeToken, 280) ->
 283;
yeccgoto(safeToken, 281) ->
 283;
yeccgoto(safeToken, 284) ->
 283;
yeccgoto(safeToken, 287) ->
 187;
yeccgoto(safeToken, 291) ->
 187;
yeccgoto(safeToken, 317) ->
 187;
yeccgoto(safeToken, 324) ->
 187;
yeccgoto(safeToken, 338) ->
 237;
yeccgoto(safeToken, 347) ->
 187;
yeccgoto(safeToken, 353) ->
 187;
yeccgoto(safeToken, 371) ->
 373;
yeccgoto(safeToken, 374) ->
 187;
yeccgoto(safeToken, 377) ->
 187;
yeccgoto(safeToken, 382) ->
 393;
yeccgoto(safeToken, 395) ->
 393;
yeccgoto(safeToken, 409) ->
 373;
yeccgoto(safeToken, 411) ->
 187;
yeccgoto(safeToken, 414) ->
 187;
yeccgoto(safeToken, 419) ->
 393;
yeccgoto(safeToken, 427) ->
 393;
yeccgoto(safeToken, 439) ->
 187;
yeccgoto(safeToken, 443) ->
 187;
yeccgoto(safeToken, 445) ->
 393;
yeccgoto(safeToken, 449) ->
 393;
yeccgoto(safeToken, 495) ->
 187;
yeccgoto(safeToken, 498) ->
 187;
yeccgoto(safeToken, 506) ->
 259;
yeccgoto(safeToken, 508) ->
 187;
yeccgoto(safeToken, 511) ->
 513;
yeccgoto(safeToken, 525) ->
 527;
yeccgoto(safeToken, 529) ->
 237;
yeccgoto(safeToken, 534) ->
 536;
yeccgoto(safeToken, 538) ->
 373;
yeccgoto(safeToken, 540) ->
 187;
yeccgoto(safeToken, 544) ->
 187;
yeccgoto(safeToken, 547) ->
 393;
yeccgoto(safeToken, 554) ->
 237;
yeccgoto(safeToken, 567) ->
 158;
yeccgoto(safeToken, 571) ->
 158;
yeccgoto(safeToken, 576) ->
 158;
yeccgoto(safeToken, 584) ->
 158;
yeccgoto(safeToken, 590) ->
 158;
yeccgoto(safeToken, 596) ->
 607;
yeccgoto(safeToken, 617) ->
 607;
yeccgoto(safeToken, 623) ->
 624;
yeccgoto(safeToken, 625) ->
 628;
yeccgoto(safeToken, 629) ->
 216;
yeccgoto(safeToken, 631) ->
 632;
yeccgoto(safeToken, 635) ->
 636;
yeccgoto(safeToken, 637) ->
 638;
yeccgoto(safeToken, 639) ->
 640;
yeccgoto(safeToken, 641) ->
 158;
yeccgoto(safeToken, 649) ->
 373;
yeccgoto(safeToken, 668) ->
 158;
yeccgoto(safeToken, 671) ->
 158;
yeccgoto(safeToken, 684) ->
 685;
yeccgoto(safeToken, 687) ->
 685;
yeccgoto(safeToken, 692) ->
 675;
yeccgoto(safeToken, 706) ->
 131;
yeccgoto(safeToken, 731) ->
 158;
yeccgoto(safeToken, 760) ->
 187;
yeccgoto(safeToken, 763) ->
 187;
yeccgoto(safeToken, 768) ->
 216;
yeccgoto(safeToken, 770) ->
 513;
yeccgoto(safeToken, 772) ->
 513;
yeccgoto(safeToken, 778) ->
 158;
yeccgoto(safeToken, 803) ->
 158;
yeccgoto(safeToken, 809) ->
 158;
yeccgoto(safeToken, 816) ->
 158;
yeccgoto(safeToken, 822) ->
 824;
yeccgoto(safeToken, 829) ->
 158;
yeccgoto(safeToken, 834) ->
 675;
yeccgoto(safeToken, 838) ->
 840;
yeccgoto(safeToken, 840) ->
 840;
yeccgoto(safeToken, 843) ->
 628;
yeccgoto(safeToken, 849) ->
 628;
yeccgoto(safeToken, 853) ->
 854;
yeccgoto(safeToken, 855) ->
 856;
yeccgoto(safeToken, 857) ->
 858;
yeccgoto(secondEventParameter, 419) ->
 426;
yeccgoto(secondEventParameter, 427) ->
 430;
yeccgoto(secondEventParameters, 426) ->
 428;
yeccgoto(secondEventParameters, 430) ->
 431;
yeccgoto(secondRequestedEvent, 411) ->
 413;
yeccgoto(secondRequestedEvent, 414) ->
 417;
yeccgoto(secondRequestedEventBody, 412) ->
 420;
yeccgoto(secondRequestedEvents, 413) ->
 415;
yeccgoto(secondRequestedEvents, 417) ->
 418;
yeccgoto(servChgReplyParm, 787) ->
 792;
yeccgoto(servChgReplyParm, 798) ->
 801;
yeccgoto(servChgReplyParms, 792) ->
 799;
yeccgoto(servChgReplyParms, 801) ->
 802;
yeccgoto(serviceChangeAddress, 596) ->
 608;
yeccgoto(serviceChangeAddress, 617) ->
 608;
yeccgoto(serviceChangeAddress, 787) ->
 793;
yeccgoto(serviceChangeAddress, 798) ->
 793;
yeccgoto(serviceChangeDelay, 596) ->
 609;
yeccgoto(serviceChangeDelay, 617) ->
 609;
yeccgoto(serviceChangeDescriptor, 592) ->
 594;
yeccgoto(serviceChangeMethod, 596) ->
 610;
yeccgoto(serviceChangeMethod, 617) ->
 610;
yeccgoto(serviceChangeMgcId, 596) ->
 611;
yeccgoto(serviceChangeMgcId, 617) ->
 611;
yeccgoto(serviceChangeMgcId, 787) ->
 794;
yeccgoto(serviceChangeMgcId, 798) ->
 794;
yeccgoto(serviceChangeParm, 596) ->
 612;
yeccgoto(serviceChangeParm, 617) ->
 620;
yeccgoto(serviceChangeParms, 612) ->
 618;
yeccgoto(serviceChangeParms, 620) ->
 621;
yeccgoto(serviceChangeProfile, 596) ->
 613;
yeccgoto(serviceChangeProfile, 617) ->
 613;
yeccgoto(serviceChangeProfile, 787) ->
 795;
yeccgoto(serviceChangeProfile, 798) ->
 795;
yeccgoto(serviceChangeReason, 596) ->
 614;
yeccgoto(serviceChangeReason, 617) ->
 614;
yeccgoto(serviceChangeReply, 708) ->
 725;
yeccgoto(serviceChangeReply, 726) ->
 725;
yeccgoto(serviceChangeReplyBody, 779) ->
 781;
yeccgoto(serviceChangeReplyDescriptor, 780) ->
 784;
yeccgoto(serviceChangeRequest, 132) ->
 154;
yeccgoto(serviceChangeRequest, 562) ->
 154;
yeccgoto(serviceChangeVersion, 596) ->
 615;
yeccgoto(serviceChangeVersion, 617) ->
 615;
yeccgoto(serviceChangeVersion, 787) ->
 796;
yeccgoto(serviceChangeVersion, 798) ->
 796;
yeccgoto(serviceState, 329) ->
 333;
yeccgoto(serviceStates, 317) ->
 322;
yeccgoto(serviceStates, 324) ->
 322;
yeccgoto(sigParameter, 197) ->
 204;
yeccgoto(sigParameter, 205) ->
 208;
yeccgoto(sigParameters, 204) ->
 206;
yeccgoto(sigParameters, 208) ->
 209;
yeccgoto(signalList, 184) ->
 188;
yeccgoto(signalList, 192) ->
 188;
yeccgoto(signalListId, 258) ->
 260;
yeccgoto(signalListId, 506) ->
 507;
yeccgoto(signalListParm, 261) ->
 262;
yeccgoto(signalListParm, 264) ->
 267;
yeccgoto(signalListParm, 508) ->
 509;
yeccgoto(signalListParms, 262) ->
 265;
yeccgoto(signalListParms, 267) ->
 268;
yeccgoto(signalName, 184) ->
 189;
yeccgoto(signalName, 192) ->
 189;
yeccgoto(signalName, 261) ->
 189;
yeccgoto(signalName, 264) ->
 189;
yeccgoto(signalName, 498) ->
 189;
yeccgoto(signalName, 508) ->
 189;
yeccgoto(signalParm, 184) ->
 190;
yeccgoto(signalParm, 192) ->
 195;
yeccgoto(signalParms, 190) ->
 193;
yeccgoto(signalParms, 195) ->
 196;
yeccgoto(signalRequest, 184) ->
 191;
yeccgoto(signalRequest, 192) ->
 191;
yeccgoto(signalRequest, 261) ->
 263;
yeccgoto(signalRequest, 264) ->
 263;
yeccgoto(signalRequest, 498) ->
 504;
yeccgoto(signalRequest, 508) ->
 263;
yeccgoto(signalType, 239) ->
 243;
yeccgoto(signalsDescriptor, 160) ->
 178;
yeccgoto(signalsDescriptor, 179) ->
 178;
yeccgoto(signalsDescriptor, 400) ->
 403;
yeccgoto(signalsDescriptor, 432) ->
 433;
yeccgoto(signalsDescriptor, 733) ->
 752;
yeccgoto(signalsDescriptor, 756) ->
 752;
yeccgoto(signalsDescriptor, 813) ->
 752;
yeccgoto(statisticsDescriptor, 733) ->
 753;
yeccgoto(statisticsDescriptor, 756) ->
 753;
yeccgoto(statisticsDescriptor, 813) ->
 753;
yeccgoto(statisticsParameter, 760) ->
 762;
yeccgoto(statisticsParameter, 763) ->
 766;
yeccgoto(statisticsParameters, 762) ->
 764;
yeccgoto(statisticsParameters, 766) ->
 767;
yeccgoto(streamDescriptor, 301) ->
 309;
yeccgoto(streamDescriptor, 312) ->
 309;
yeccgoto(streamID, 236) ->
 238;
yeccgoto(streamID, 338) ->
 339;
yeccgoto(streamID, 529) ->
 530;
yeccgoto(streamID, 554) ->
 555;
yeccgoto(streamModes, 364) ->
 370;
yeccgoto(streamParm, 301) ->
 310;
yeccgoto(streamParm, 312) ->
 310;
yeccgoto(streamParm, 340) ->
 341;
yeccgoto(streamParm, 342) ->
 345;
yeccgoto(streamParmList, 341) ->
 343;
yeccgoto(streamParmList, 345) ->
 346;
yeccgoto(subtractRequest, 132) ->
 155;
yeccgoto(subtractRequest, 562) ->
 155;
yeccgoto(terminationA, 567) ->
 568;
yeccgoto(terminationA, 571) ->
 568;
yeccgoto(terminationAudit, 733) ->
 754;
yeccgoto(terminationAudit, 813) ->
 814;
yeccgoto(terminationB, 576) ->
 577;
yeccgoto(terminationID, 157) ->
 159;
yeccgoto(terminationID, 272) ->
 274;
yeccgoto(terminationID, 275) ->
 278;
yeccgoto(terminationID, 567) ->
 569;
yeccgoto(terminationID, 571) ->
 569;
yeccgoto(terminationID, 576) ->
 578;
yeccgoto(terminationID, 584) ->
 585;
yeccgoto(terminationID, 590) ->
 591;
yeccgoto(terminationID, 641) ->
 642;
yeccgoto(terminationID, 668) ->
 669;
yeccgoto(terminationID, 671) ->
 672;
yeccgoto(terminationID, 731) ->
 732;
yeccgoto(terminationID, 778) ->
 779;
yeccgoto(terminationID, 803) ->
 804;
yeccgoto(terminationID, 809) ->
 812;
yeccgoto(terminationID, 816) ->
 274;
yeccgoto(terminationID, 829) ->
 812;
yeccgoto(terminationIDList, 270) ->
 273;
yeccgoto(terminationIDList, 810) ->
 818;
yeccgoto(terminationIDList, 830) ->
 818;
yeccgoto(terminationIDListRepeat, 274) ->
 276;
yeccgoto(terminationIDListRepeat, 278) ->
 279;
yeccgoto(terminationStateDescriptor, 301) ->
 311;
yeccgoto(terminationStateDescriptor, 312) ->
 311;
yeccgoto(terminationStateParm, 317) ->
 323;
yeccgoto(terminationStateParm, 324) ->
 327;
yeccgoto(terminationStateParms, 323) ->
 325;
yeccgoto(terminationStateParms, 327) ->
 328;
yeccgoto(timeStamp, 435) ->
 440;
yeccgoto(timeStamp, 456) ->
 440;
yeccgoto(timeStamp, 596) ->
 616;
yeccgoto(timeStamp, 617) ->
 616;
yeccgoto(timeStamp, 651) ->
 440;
yeccgoto(timeStamp, 653) ->
 440;
yeccgoto(timeStamp, 787) ->
 797;
yeccgoto(timeStamp, 798) ->
 797;
yeccgoto(topologyDescriptor, 132) ->
 156;
yeccgoto(topologyDescriptor, 562) ->
 156;
yeccgoto(topologyDescriptor, 708) ->
 156;
yeccgoto(topologyDescriptor, 726) ->
 156;
yeccgoto(topologyDirection, 579) ->
 583;
yeccgoto(topologyTriple, 567) ->
 570;
yeccgoto(topologyTriple, 571) ->
 574;
yeccgoto(topologyTripleList, 570) ->
 572;
yeccgoto(topologyTripleList, 574) ->
 575;
yeccgoto(transactionAck, 684) ->
 686;
yeccgoto(transactionAck, 687) ->
 690;
yeccgoto(transactionAckList, 686) ->
 688;
yeccgoto(transactionAckList, 690) ->
 691;
yeccgoto(transactionID, 120) ->
 676;
yeccgoto(transactionID, 692) ->
 693;
yeccgoto(transactionID, 834) ->
 835;
yeccgoto(transactionItem, 97) ->
 113;
yeccgoto(transactionItem, 113) ->
 113;
yeccgoto(transactionList, 97) ->
 114;
yeccgoto(transactionList, 113) ->
 119;
yeccgoto(transactionPending, 97) ->
 115;
yeccgoto(transactionPending, 113) ->
 115;
yeccgoto(transactionReply, 97) ->
 116;
yeccgoto(transactionReply, 113) ->
 116;
yeccgoto(transactionReplyBody, 696) ->
 700;
yeccgoto(transactionRequest, 97) ->
 117;
yeccgoto(transactionRequest, 113) ->
 117;
yeccgoto(transactionResponseAck, 97) ->
 118;
yeccgoto(transactionResponseAck, 113) ->
 118;
yeccgoto(value, 210) ->
 223;
yeccgoto(value, 211) ->
 219;
yeccgoto(value, 212) ->
 218;
yeccgoto(value, 213) ->
 217;
yeccgoto(value, 220) ->
 233;
yeccgoto(value, 221) ->
 224;
yeccgoto(value, 225) ->
 231;
yeccgoto(value, 226) ->
 229;
yeccgoto(value, 629) ->
 630;
yeccgoto(value, 768) ->
 769;
yeccgoto(valueList, 224) ->
 227;
yeccgoto(valueList, 229) ->
 230;
yeccgoto(valueList, 233) ->
 234;
yeccgoto(__Symbol, __State) ->
 exit({__Symbol, __State, missing_in_goto_table}).


